@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include "HttpResponseMetadata.h"
 #include "HttpRequestLine.h"
 #include "RequestPtr.h"
+#include "RequestTag.h"
 #include "ErrorResponse.h"
 #include "ErrorDescriptor.h"
 #include "ErrorCodes.h"
@@ -45,10 +46,10 @@
 #include "HttpStreamWriter.h"
 #include "HttpMessageShaper.h"
 #include "HttpVersion.h"
+#include "TimeDelta.h"
 #include "URI.h"
 #include "NetworkActivityReporter.h"
 #include "Debug.h"
-#include <ace/Time_Value.h>
 #include <stddef.h>
 #include <cassert>
 
@@ -64,9 +65,8 @@ class Client::ResponseHandler :
 	DECLARE_NON_COPYABLE(ResponseHandler)
 public:
 	ResponseHandler(
-		ClientWeakPtr const& client_ptr,
-		ConstRequestPtr const& request,
-		long id, bool conn_close);
+		ClientWeakPtr const& client_ptr, ConstRequestPtr const& request,
+		RequestTag const& request_tag, long id, bool conn_close);
 	
 	virtual ~ResponseHandler();
 private:
@@ -74,9 +74,9 @@ private:
 	
 	virtual void processResponseMetadata(std::auto_ptr<HttpResponseMetadata> metadata);
 	
-	virtual void processBodyData(SplittableBuffer& data, bool eof, DownloadProgress const& progress);
+	virtual void processBodyData(SplittableBuffer& data, bool eof);
 	
-	virtual void processError(std::auto_ptr<ErrorDescriptor> edesc);
+	virtual void processError(IntrusivePtr<ErrorDescriptor> const& edesc);
 	
 	virtual void processNewData(SplittableBuffer& data, bool eof);
 	
@@ -84,6 +84,7 @@ private:
 	
 	ClientWeakPtr m_ptrClient;
 	ConstRequestPtr const m_ptrRequest;
+	RequestTag m_requestTag;
 	long const m_id;
 	bool m_isConnClose;
 	bool m_isEOF;
@@ -112,19 +113,19 @@ Client::Client(ServiceContext& context, SocketPtr const& client_stream)
 	m_ptrClientStream(client_stream),
 	m_isEndOfOutput(false),
 	m_readTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Client::onReadTimeout),
-		ACE_Time_Value(READ_TIMEOUT)
+		TimeDelta::fromSec(READ_TIMEOUT),
+		&context.reactor()
 	),
 	m_writeTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Client::onWriteTimeout),
-		ACE_Time_Value(WRITE_TIMEOUT)
+		TimeDelta::fromSec(WRITE_TIMEOUT),
+		&context.reactor()
 	),
 	m_idleTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Client::onIdleTimeout),
-		ACE_Time_Value(IDLE_TIMEOUT)
+		TimeDelta::fromSec(IDLE_TIMEOUT),
+		&context.reactor()
 	),
 	m_eventMask(Reactor::READ),
 	m_flags(0),
@@ -236,7 +237,7 @@ Client::processInputData(bool eof)
 		if (isErrorState()) {
 			DBG_INCOMING_REQUEST_ERROR();
 			//DEBUGLOG("ERROR parsing request");
-			auto_ptr<ErrorDescriptor> edesc(errorDescriptor());
+			IntrusivePtr<ErrorDescriptor> edesc(errorDescriptor());
 			if (!edesc.get()) {
 				edesc.reset(new ErrorDescriptor(
 					ErrorCodes::ERROR_PARSING_RESPONSE,
@@ -264,16 +265,18 @@ Client::processInputData(bool eof)
 				HttpRequestLine rline(BString("GET"), URI(BString()), HttpVersion::HTTP_1_0);
 				RequestPtr request(new HttpRequestMetadata(rline));
 				request->setBodyStatus(HttpRequestMetadata::BODY_FORBIDDEN);
+				RequestTag request_tag;
 				IntrusivePtr<ResponseHandler> response_handler(
 					new ResponseHandler(
-						m_ptrSelf, request,
+						m_ptrSelf, request, request_tag,
 						++m_responseIdCounter, true /* conn_close */
 					)
 				);
 				ScopedSuspender suspender(*this);
 				IntrusivePtr<AbstractRequestHandler> request_handler(
 					m_rContext.server().submitRequest(
-						request, response_handler, edesc
+						request, request_tag,
+						response_handler, edesc
 					)
 				);
 				SplittableBuffer body_data;
@@ -461,15 +464,16 @@ Client::eventGotMetadata(auto_ptr<HttpRequestMetadata> metadata, bool is_persist
 	ConstRequestPtr request(metadata.release());
 	// metadata is null now !!!
 	
+	RequestTag request_tag;
 	IntrusivePtr<ResponseHandler> response_handler(
 		new ResponseHandler(
-			m_ptrSelf, request,
+			m_ptrSelf, request, request_tag,
 			++m_responseIdCounter, !is_persistent
 		)
 	);
 	ScopedSuspender suspender(*this);
 	m_ptrRequestHandler = m_rContext.server().submitRequest(
- 		request, response_handler
+ 		request, request_tag, response_handler
 	);
 }
 
@@ -526,11 +530,11 @@ Client::handleWrite(ACE_HANDLE)
 /*======================== Client::ResponseHandler ===========================*/
 
 Client::ResponseHandler::ResponseHandler(
-	ClientWeakPtr const& client_ptr,
-	ConstRequestPtr const& request,
-	long id, bool conn_close)
+	ClientWeakPtr const& client_ptr, ConstRequestPtr const& request,
+	RequestTag const& request_tag, long id, bool conn_close)
 :	m_ptrClient(client_ptr),
 	m_ptrRequest(request),
+	m_requestTag(request_tag),
 	m_id(id),
 	m_isConnClose(conn_close),
 	m_isEOF(false),
@@ -570,11 +574,11 @@ Client::ResponseHandler::processResponseMetadata(std::auto_ptr<HttpResponseMetad
 	typedef AbstractDebugAgent ADA;
 	ADA::HttpMessageType msg_type = ADA::UNMODIFIED_OUTGOING_RESPONSE;
 	int const scode = metadata->statusLine().getCode();
-	if (scode >= 400 && scode < 600) {
-		msg_type = ADA::ERROR_OUTGOING_RESPONSE;
-	} else if (metadata->isFlagSet(HttpResponseMetadata::CONTENT_CRAFTED)) {
+	if (m_requestTag->flags().isSet(RequestTag::RESPONSE_CRAFTED)) {
 		msg_type = ADA::CRAFTED_OUTGOING_RESPONSE;
-	} else if (metadata->isFlagSet(HttpResponseMetadata::CONTENT_MODIFIED)) {
+	} else if (scode >= 400 && scode < 600) {
+		msg_type = ADA::ERROR_OUTGOING_RESPONSE;
+	} else if (m_requestTag->flags().isSet(RequestTag::RESPONSE_MODIFIED)) {
 		msg_type = ADA::FILTERED_OUTGOING_RESPONSE;
 	}
 	DBG_HTTP_MESSAGE_BEGIN(*metadata, msg_type);
@@ -582,7 +586,7 @@ Client::ResponseHandler::processResponseMetadata(std::auto_ptr<HttpResponseMetad
 }
 
 void
-Client::ResponseHandler::processBodyData(SplittableBuffer& data, bool eof, DownloadProgress const&)
+Client::ResponseHandler::processBodyData(SplittableBuffer& data, bool eof)
 {
 	if (m_isEOF) {
 		data.clear();
@@ -596,7 +600,7 @@ Client::ResponseHandler::processBodyData(SplittableBuffer& data, bool eof, Downl
 }
 
 void
-Client::ResponseHandler::processError(auto_ptr<ErrorDescriptor> edesc)
+Client::ResponseHandler::processError(IntrusivePtr<ErrorDescriptor> const& edesc)
 {
 	DEBUGLOG(
 		"Client::ResponseHandler::processError(): "
