@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -27,6 +27,7 @@
 #include "URI.h"
 #include "HttpRequestMetadata.h"
 #include "HttpRequestLine.h"
+#include "HttpStatusLine.h"
 #include "AbstractRequestHandler.h"
 #include "AbstractResponseHandler.h"
 #include "FilterTryList.h"
@@ -39,10 +40,12 @@
 #include "ReplacementJs.h"
 #include "RegexFilterDescriptor.h"
 #include "SubstitutionRequestParser.h"
+#include "State.h"
 #include "GlobalState.h"
 #include "UrlPatterns.h"
 #include "ContentFilters.h"
 #include "BString.h"
+#include "RequestTag.h"
 #include <string>
 
 using namespace std;
@@ -91,24 +94,23 @@ FilteringServer::~FilteringServer()
 IntrusivePtr<AbstractRequestHandler>
 FilteringServer::submitRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler)
 {
 	IntrusivePtr<AbstractRequestHandler> res;
-	if ((res = processSubstitutionRequest(request_metadata, handler))) {
+	if ((res = processSubstitutionRequest(request_metadata, request_tag, handler))) {
 		// do nothing
-	} else if ((res = processUrlPatterns(request_metadata, handler))) {
+	} else if ((res = processUrlPatterns(request_metadata, request_tag, handler))) {
 		// do nothing
-	} else if ((res = processAnalyzeRequest(request_metadata, handler))) {
+	} else if ((res = processAnalyzeRequest(request_metadata, request_tag, handler))) {
 		// do nothing
 	} else {
+		URI const& uri = request_metadata->requestLine().getURI();
 		std::auto_ptr<FilterTryList> filters(new FilterTryList);
 		{
 			GlobalState::ReadAccessor global_state; // hold the lock
-			if (global_state->isFilteringEnabled()) {
-				applyStandardFilters(
-					*filters,
-					request_metadata->requestLine().getURI()
-				);
+			if (isFilteringEnabledForURL(*global_state, uri)) {
+				applyStandardFilters(*filters, uri);
 			}
 			if (global_state->config().isClientCompressionEnabled()) {
 				filters->append(FilterTryList::FilterConstructorPtr(
@@ -118,7 +120,7 @@ FilteringServer::submitRequest(
 				));
 			}
 		}
-		res = Server::submitRequest(request_metadata, handler, filters);
+		res = Server::submitRequest(request_metadata, request_tag, handler, filters);
 	}
 	return res;
 }
@@ -126,6 +128,7 @@ FilteringServer::submitRequest(
 IntrusivePtr<AbstractRequestHandler>
 FilteringServer::processSubstitutionRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler)
 {
 	BString const si_prefix("bf-si-");
@@ -136,7 +139,7 @@ FilteringServer::processSubstitutionRequest(
 	ImageSubstitutionParser image_parser(is_head_request);
 	if (image_parser.parse(uri, si_prefix)) {
 		return Server::submitRequest(
-			request_metadata, handler,
+			request_metadata, request_tag, handler,
 			image_parser.responseGenerator()
 		);
 	}
@@ -144,7 +147,7 @@ FilteringServer::processSubstitutionRequest(
 	FlashSubstitutionParser flash_parser(is_head_request);
 	if (flash_parser.parse(uri, sf_prefix)) {
 		return Server::submitRequest(
-			request_metadata, handler,
+			request_metadata, request_tag, handler,
 			flash_parser.responseGenerator()
 		);
 	}
@@ -155,71 +158,79 @@ FilteringServer::processSubstitutionRequest(
 IntrusivePtr<AbstractRequestHandler>
 FilteringServer::processUrlPatterns(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler)
 {
 	URI const& url = request_metadata->requestLine().getURI();
-	GlobalState::ReadAccessor global_state;
-	if (global_state->urlPatterns().isForbidden(url)) {
-		return Server::submitRequest(
-			request_metadata, handler,
-			sigc::bind(
+	bool const is_head_request = request_metadata->requestLine().getMethod() == BString("HEAD");
+	UrlPatterns::Substitution subst = GlobalState::ReadAccessor()->urlPatterns().getSubstitutionFor(url);
+	CraftedResponseGenerator rgen;
+	
+	if (subst == UrlPatterns::AUTO_SUBST) {
+		BString const& path = request_metadata->requestLine().getURI().getRawPath();
+		BString const swf(".swf");
+		if (StringUtils::ciEndsWith(path.begin(), path.end(), swf.begin(), swf.end())) {
+			subst = UrlPatterns::EMPTY_FLASH;
+		} else {
+			subst = UrlPatterns::EMPTY_IMAGE;
+		}
+	}
+	
+	switch (subst) {
+		case UrlPatterns::FORBID_HTML: {
+			rgen = sigc::bind(
 				sigc::ptr_fun(&ErrorResponse::errUrlForbidden),
 				url.toString()
-			)
-		);
-	}
-	
-	bool const is_head_request = request_metadata->requestLine().getMethod() == BString("HEAD");
-	UrlPatterns::Substitution subst = global_state->urlPatterns().getSubstitutionFor(url);
-	switch (subst) {
-		case UrlPatterns::EMPTY_IMAGE: {
-			return Server::submitRequest(
-				request_metadata, handler,
-				sigc::bind(
-					sigc::ptr_fun(&ReplacementImage::createHttpResponse),
-					is_head_request,
-					1, 1 // width, height
-				)
 			);
+			break;
+		}
+		case UrlPatterns::EMPTY_IMAGE: {
+			rgen = sigc::bind(
+				sigc::ptr_fun(&ReplacementImage::createHttpResponse),
+				is_head_request,
+				1, 1, // width, height
+				int(HttpStatusLine::SC_FORBIDDEN)
+			);
+			break;
 		}
 		case UrlPatterns::EMPTY_FLASH: {
-			return Server::submitRequest(
-				request_metadata, handler,
-				sigc::bind(
-					sigc::ptr_fun(&ReplacementFlash::createHttpResponse),
-					is_head_request,
-					1, 1, // width, height
-					url.toString(), ReplacementFlash::DEFAULT_FRAMERATE 
-				)
+			rgen = sigc::bind(
+				sigc::ptr_fun(&ReplacementFlash::createHttpResponse),
+				is_head_request,
+				1, 1, // width, height
+				url.toString(), ReplacementFlash::DEFAULT_FRAMERATE
 			);
+			break;
 		}
 		case UrlPatterns::EMPTY_HTML: {
-			return Server::submitRequest(
-				request_metadata, handler,
-				sigc::bind(
-					sigc::ptr_fun(&ReplacementHtml::createHttpResponse),
-					is_head_request
-				)
+			rgen = sigc::bind(
+				sigc::ptr_fun(&ReplacementHtml::createHttpResponse),
+				is_head_request
 			);
+			break;
 		}
 		case UrlPatterns::EMPTY_JS: {
-			return Server::submitRequest(
-				request_metadata, handler,
-				sigc::bind(
-					sigc::ptr_fun(&ReplacementJs::createHttpResponse),
-					is_head_request
-				)
+			rgen = sigc::bind(
+				sigc::ptr_fun(&ReplacementJs::createHttpResponse),
+				is_head_request
 			);
+			break;
+		}
+		default: {
+			return IntrusivePtr<AbstractRequestHandler>();
 		}
-		default: break;
 	}
 	
-	return IntrusivePtr<AbstractRequestHandler>();
+	return Server::submitRequest(
+		request_metadata, request_tag, handler, rgen
+	);
+	
 }
 
 IntrusivePtr<AbstractRequestHandler>
 FilteringServer::processAnalyzeRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler)
 {
 	IntrusivePtr<AbstractRequestHandler> res;
@@ -248,7 +259,7 @@ FilteringServer::processAnalyzeRequest(
 		RequestPtr req(new HttpRequestMetadata(*request_metadata));
 		req->requestLine().setURI(new_uri);
 		std::auto_ptr<FilterTryList> filter_try_list(new FilterTryList);
-		if (GlobalState::ReadAccessor()->isFilteringEnabled()) {
+		if (isFilteringEnabledForURL(*GlobalState::ReadAccessor(), new_uri)) {
 			filter_try_list->append(FilterTryList::FilterConstructorPtr(
 				new FilterTryList::FilterConstructor(
 					sigc::bind(
@@ -257,8 +268,9 @@ FilteringServer::processAnalyzeRequest(
 					)
 				)
 			));
+			request_tag->flags().set(RequestTag::REQUEST_ANALYZE);
 		}
-		res = Server::submitRequest(req, handler, filter_try_list);
+		res = Server::submitRequest(req, request_tag, handler, filter_try_list);
 	}
 	
 	return res;
@@ -267,7 +279,20 @@ FilteringServer::processAnalyzeRequest(
 void
 FilteringServer::applyStandardFilters(FilterTryList& filters, URI const& request_url)
 {
-	BString url_str(request_url.toBString());
+	BString const url_str(request_url.toBString());
+	
+	filters.append(FilterTryList::FilterConstructorPtr(
+		new FilterTryList::FilterConstructor(
+			sigc::ptr_fun(&FilterTryList::tryTextContentValidator)
+		)
+	));
+
+	// Note: HtmlResponseFilter depends on the presence of HtmlContentValidator
+	filters.append(FilterTryList::FilterConstructorPtr(
+		new FilterTryList::FilterConstructor(
+			sigc::ptr_fun(&FilterTryList::tryHtmlContentValidator)
+		)
+	));
 	
 	GlobalState::ReadAccessor state; // hold the read lock
 	
@@ -276,7 +301,7 @@ FilteringServer::applyStandardFilters(FilterTryList& filters, URI const& request
 	FilterList::const_iterator it(avail_filters.begin());
 	FilterList::const_iterator const end(avail_filters.end());
 	for (; it != end && (*it)->order() < 0; ++it) {
-		maybeApplyRegexFilter(filters, url_str, *it);
+		filters.tryAppendRegexFilter(url_str, *it);
 	}
 	filters.append(FilterTryList::FilterConstructorPtr(
 		new FilterTryList::FilterConstructor(
@@ -284,25 +309,18 @@ FilteringServer::applyStandardFilters(FilterTryList& filters, URI const& request
 		)
 	));
 	for (; it != end; ++it) {
-		maybeApplyRegexFilter(filters, url_str, *it);
+		filters.tryAppendRegexFilter(url_str, *it);
 	}
 }
 
-void
-FilteringServer::maybeApplyRegexFilter(
-	FilterTryList& filters, BString const& request_url,
-	IntrusivePtr<RegexFilterDescriptor const> const& flt)
+bool
+FilteringServer::isFilteringEnabledForURL(State const& state, URI const& url)
 {
-	if (flt->isValid() && flt->urlMatches(request_url)) {
-		filters.append(FilterTryList::FilterConstructorPtr(
-			new FilterTryList::FilterConstructor(
-				sigc::bind(
-					sigc::ptr_fun(&FilterTryList::tryRegexFilter),
-					flt
-				)
-			)
-		));
+	if (!state.isFilteringEnabled()) {
+		return false;
 	}
+	
+	return !state.urlPatterns().isFilteringDisabled(url);
 }
 
 
@@ -313,8 +331,8 @@ FilteringServer::ImageSubstitutionParser::handleResult(
 {
 	m_responseGenerator = sigc::bind(
 		sigc::ptr_fun(&ReplacementImage::createHttpResponse),
-		isHeadResponse(),
-		width, height
+		isHeadResponse(), width, height,
+		int(HttpStatusLine::SC_OK)
 	);
 }
 
