@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 
 #include "HtmlProcessor.h"
 #include "HtmlLexerDefinitions.h"
+#include "NonCopyable.h"
 #include "ServiceContext.h"
 #include "HtmlElementNode.h"
 #include "HtmlTextNode.h"
@@ -65,7 +66,7 @@
 #include <stddef.h>
 #include <limits>
 #include <memory>
-#include <deque>
+#include <vector>
 #include <algorithm>
 #include <sstream>
 #include <cstring>
@@ -144,10 +145,12 @@ public:
 	
 	int getNestLevel() const { return m_nestLevel; }
 	
-	int getNestLevel() { return m_nestLevel; }
-	
 	void setNestLevel(int level) { m_nestLevel = level; }
 	
+	int getInnerHtmlLevel() const { return m_innerHtmlLevel; }
+	
+	void setInnerHtmlLevel(int level) { m_innerHtmlLevel = level; }
+	
 	int getDescendantScriptFetches() const { return m_descendantScriptFetches; }
 	
 	int incDescendantScriptFetches() { return ++m_descendantScriptFetches; }
@@ -156,6 +159,10 @@ public:
 	
 	void setExternalScriptInvolved() { m_flags |= EXTERNAL; }
 	
+	bool isInnerHtmlInvolved() const { return m_flags & INNER_HTML; }
+	
+	void setInnerHtmlInvolved() { m_flags |= INNER_HTML; }
+	
 	bool isRedirectInvolved() const { return m_flags & REDIRECT; }
 	
 	void setRedirectInvolved() { m_flags |= REDIRECT; }
@@ -216,7 +223,8 @@ private:
 		EXTERNAL = 1, REDIRECT = 2, COOKIE = 4,
 		CSS = 8, ONLOAD_ASSIGNED = 16,
 		DISCARD_SUBPROCESSOR_OUTPUT = 32,
-		TOO_MANY_UNRELATED_SCRIPTS = 64
+		TOO_MANY_UNRELATED_SCRIPTS = 64,
+		INNER_HTML = 128
 	};
 	enum {
 		WEBBUG_TAIL_EV1 = 1, WEBBUG_TAIL_EV2 = 2,
@@ -224,11 +232,14 @@ private:
 	};
 	static size_t const SCRIPT_OUTPUT_LIMIT = 2048;
 	
-	virtual void processJsOutput(char const* data, size_t length);
+	virtual void processJsOutput(char const* data, bool newline);
 	
 	virtual void processOnLoadAssignment();
 	
+	virtual void processInnerHtmlAssignment(char const* data);
+	
 	int m_nestLevel;
+	int m_innerHtmlLevel;
 	int m_flags;
 	int m_descendantScriptFetches;
 	int m_unclickableImageCount;
@@ -241,31 +252,26 @@ private:
 	AdSuspectList m_adSuspects;
 	SplittableBuffer m_subprocessorOutput;
 	size_t m_subprocessorOutputSize;
-	std::deque<URI> m_unrelatedNestedScripts;
+	std::vector<URI> m_unrelatedNestedScripts;
 	// Urls of scripts from domains unrelated to the originating page.
 	// The original (nest level = 0) script is also present here.
-	JsEnvironment::ListenerHandle m_listenerHandle;
 };
 
 
 class HtmlProcessor::SubProcessor : public HtmlProcessor
 {
+	DECLARE_NON_COPYABLE(SubProcessor)
 public:
-	SubProcessor(HtmlProcessor& parent, ScriptContext& context)
-	: HtmlProcessor(parent, context), m_rContext(context),
-	  m_oldNestLevel(context.getNestLevel()),
-	  m_pOldSubProcessor(context.getSubProcessor()) {
-		m_rContext.setNestLevel(m_oldNestLevel + 1);
-		m_rContext.setSubProcessor(this);
-	}
+	enum Type { DOCUMENT_WRITE, INNER_HTML };
 	
-	~SubProcessor() {
-		m_rContext.setSubProcessor(m_pOldSubProcessor);
-		m_rContext.setNestLevel(m_oldNestLevel);
-	}
+	SubProcessor(HtmlProcessor& parent, ScriptContext& context, Type type);
+	
+	~SubProcessor();
 private:
 	ScriptContext& m_rContext;
+	Type m_type;
 	int m_oldNestLevel;
+	int m_oldInnerHtmlLevel;
 	HtmlProcessor* m_pOldSubProcessor;
 };
 
@@ -284,7 +290,7 @@ class HtmlProcessor::LeaveAsIsScriptOperation : public AbstractScriptOperation
 public:
 	LeaveAsIsScriptOperation(
 		HtmlProcessor& processor, HtmlNode* node,
-		SplittableBuffer const& code, BString const& comment);
+		SplittableBuffer const* code, BString const& comment);
 	
 	virtual ~LeaveAsIsScriptOperation();
 	
@@ -292,7 +298,7 @@ public:
 private:
 	HtmlProcessor& m_rProcessor;
 	HtmlNode* m_pNode;
-	SplittableBuffer const& m_rCode;
+	SplittableBuffer const* m_pCode;
 	BString m_comment;
 };
 
@@ -332,6 +338,49 @@ private:
 };
 
 
+class HtmlProcessor::TempStreamScope
+{
+public:
+	TempStreamScope(SBOutStream& strm) : m_rStream(strm) {
+		assert(m_rStream.data().empty() && "TempStreamScope's can't be nested");
+	}
+	
+	~TempStreamScope() { m_rStream.clearData(); }
+private:
+	SBOutStream& m_rStream;
+};
+
+
+struct HtmlProcessor::ScriptFetchResult
+{
+	ScriptFetchResult()
+	: numRedirects(0), persistentCookies(false), dynamicContent(false) {}
+	
+	IntrusivePtr<ErrorDescriptor> errorDescriptor;
+	int numRedirects;
+	bool persistentCookies;
+	bool dynamicContent;
+};
+
+
+class HtmlProcessor::PageOpenListener : public JsEnvironment::PageOpenListener
+{
+public:
+	PageOpenListener(JsEnvironment& js_env);
+	
+	virtual ~PageOpenListener();
+	
+	virtual void processPageOpening(char const* url, char const* target);
+	
+	BString const& getURL() const { return m_url; }
+	
+	BString const& getTarget() const { return m_target; }
+private:
+	BString m_url;
+	BString m_target;
+};
+
+
 SplittableBuffer const HtmlProcessor::m_sEmptyBuffer;
 
 inline void
@@ -368,7 +417,9 @@ HtmlProcessor::HtmlProcessor(ServiceContext& service_context)
 	m_pContainingNode(&m_docFragment),
 	m_openingTagType(TAG_OTHER),
 	m_closingTagType(TAG_OTHER),
-	m_noscriptFollowsScript(false)
+	m_isInsideNoscript(false),
+	m_noscriptFollowsScript(false),
+	m_tempStream(4000)
 {
 }
 
@@ -389,7 +440,9 @@ HtmlProcessor::HtmlProcessor(HtmlProcessor const& parent, ScriptContext& script_
 	m_imgTagsUsingMap(parent.m_imgTagsUsingMap),
 	m_openingTagType(TAG_OTHER),
 	m_closingTagType(TAG_OTHER),
-	m_noscriptFollowsScript(parent.m_noscriptFollowsScript)
+	m_isInsideNoscript(parent.m_isInsideNoscript),
+	m_noscriptFollowsScript(parent.m_noscriptFollowsScript),
+	m_tempStream(1500) // subprocessors usually produce little output
 {
 }
 
@@ -402,16 +455,11 @@ HtmlProcessor::setRelatedInfo(ConstRequestPtr const& request, bool xhtml_content
 {
 	m_ptrRequestMetadata = request;
 	m_baseURL = request->requestLine().getURI();
+	m_jsEnv.setPageURL(m_baseURL);
 	m_contentType = (xhtml_content ? CONTENT_XHTML : CONTENT_HTML);
 	m_pageHint = HeuristicScore::getHintModifier(m_baseURL);
 }
 
-void
-HtmlProcessor::setBaseURL(URI const& url)
-{
-	m_baseURL = url;
-}
-
 void
 HtmlProcessor::consume(SplittableBuffer& data, bool eof)
 {
@@ -426,7 +474,8 @@ HtmlProcessor::reset()
 {
 	HtmlLexerCore::reset();
 	m_ptrRequestMetadata.reset(0);
-	m_baseURL = URI(BString());
+	BString empty_str;
+	URI(empty_str).swap(m_baseURL);
 	m_jsEnv.reset();
 	m_contentType = CONTENT_HTML;
 	m_ptrCurTag.reset(0);
@@ -438,8 +487,10 @@ HtmlProcessor::reset()
 	m_imgTagsUsingMap.clear();
 	m_openingTagType = TAG_OTHER;
 	m_closingTagType = TAG_OTHER;
+	m_isInsideNoscript = false;
 	m_noscriptFollowsScript = false;
 	m_processedData.clear();
+	m_tempStream.clear();
 }
 
 void
@@ -696,10 +747,10 @@ HtmlProcessor::processOpeningA(HtmlNode* anchor, bool explicit_empty)
 		URI aurl(anchor->getAttribute(BString("href")));
 		if (HeuristicScore::getUrlRelationship(aurl, m_baseURL)
 		    != HeuristicScore::URLS_UNRELATED) {
-			SBOutStream path_plus_query(150);
-			path_plus_query << aurl.getDecodedPath()
+			TempStreamScope scope(m_tempStream);
+			m_tempStream << aurl.getDecodedPath()
 				<< '?' << aurl.getDecodedQuery();
-			if (path_plus_query.data().find(BString("http://")).isAtRightBorder()) {
+			if (m_tempStream.data().find(BString("http://")).isAtRightBorder()) {
 				m_pScriptContext->incLocalLinkCount();
 			}
 		}
@@ -719,7 +770,8 @@ HtmlProcessor::processOpeningArea(HtmlNode* area, bool explicit_empty)
 bool
 HtmlProcessor::processOpeningBase(HtmlNode* base, bool explicit_empty)
 {
-	setBaseURL(URI(m_baseURL, URI(base->getAttribute(BString("href")))));
+	URI url(m_baseURL, URI(base->getAttribute(BString("href"))));
+	url.swap(m_baseURL);
 	return false;
 }
 
@@ -735,7 +787,7 @@ HtmlProcessor::processOpeningEmbed(HtmlNode* embed, bool explicit_empty)
 	InsensitiveEqual ieq;
 	
 	BString type = embed->getAttribute(BString("type"));
-	if (ieq(embed->getAttribute(BString("type")), flash_mimetype)) {
+	if (type.empty() || ieq(type, flash_mimetype)) {
 		BString src = embed->getAttribute(BString("src"));
 		URI url_rel(src);
 		URI url(m_baseURL, url_rel);
@@ -751,6 +803,7 @@ HtmlProcessor::processOpeningEmbed(HtmlNode* embed, bool explicit_empty)
 		score.processSize(width, height);
 		score.processUrl(url, url_rel, m_baseURL);
 		score.processOptions(loop, menu);
+		score.processNoscript(m_isInsideNoscript);
 		//DEBUGLOG("[score = " << score.getNumericScore() << "] EMBED: " << url);
 		if (m_pScriptContext) {
 			score.processScriptLevel(
@@ -819,6 +872,7 @@ HtmlProcessor::processOpeningMap(HtmlNode* map, bool explicit_empty)
 bool
 HtmlProcessor::processOpeningNoscript(HtmlNode* noscript, bool explicit_empty)
 {
+	m_isInsideNoscript = true;
 	if (!m_noscriptFollowsScript) {
 		return false;
 	}
@@ -894,6 +948,7 @@ HtmlProcessor::processClosingIframe(Iterator const& begin, Iterator const& end)
 	score.processPageHint(m_pageHint);
 	score.processSize(width, height);
 	score.processUrl(url, url_rel, m_baseURL);
+	score.processNoscript(m_isInsideNoscript);
 	if (m_pScriptContext) {
 		score.processScriptLevel(
 			m_pScriptContext->getNestLevel(),
@@ -933,11 +988,15 @@ HtmlProcessor::processClosingMap(Iterator const& begin, Iterator const& end)
 		typedef std::pair<MultiTagsByName::iterator, MultiTagsByName::iterator> IterPair;
 		IterPair p = m_imgTagsUsingMap.equal_range(m_ptrContainingMap->getAttribute(BString("name")));
 		for (MultiTagsByName::iterator iter = p.first; iter != p.second; ++iter) {
-			if (!iter->second->getParent()) {
+			HtmlNode* img = iter->second.get();
+			if (!img->getParent()) {
 				// this means it has been removed from the tree (together with its parent)
 				continue;
 			}
-			processImg(iter->second.get(), m_contentType==CONTENT_XHTML, anchor);
+			bool handled = processImg(img, m_contentType==CONTENT_XHTML, anchor);
+			if (!handled) {
+				processDocFragmentNode(img);
+			}
 		}
 		m_imgTagsUsingMap.erase(p.first, p.second);
 		
@@ -949,6 +1008,7 @@ HtmlProcessor::processClosingMap(Iterator const& begin, Iterator const& end)
 void
 HtmlProcessor::processClosingNoscript(Iterator const& begin, Iterator const& end)
 {
+	m_isInsideNoscript = false;
 	HtmlNode* noscript = findAncestorOrSelf(m_pContainingNode, BString("noscript"));
 	if (!noscript) {
 		processAsText(begin, end);
@@ -1072,6 +1132,7 @@ HtmlProcessor::processFlashObject(
 	score.processSize(width, height);
 	score.processUrl(url, url_rel, m_baseURL);
 	score.processOptions(loop, menu);
+	score.processNoscript(m_isInsideNoscript);
 	if (m_pScriptContext) {
 		if (node->hasAdMark()) {
 			node->getParent()->setAdMark(true);
@@ -1163,8 +1224,7 @@ HtmlProcessor::processJavaScript(HtmlNode* script, bool& nuke_adjacent_noscript)
 	BString script_src;
 	URI script_url(script_src);
 	bool const script_external = script->hasAttribute(BString("src"));
-	bool redirects = false;
-	bool cookies = false;
+	ScriptFetchResult fetch_res;
 	if (!script_external) {
 		if (script->hasChildren()) {
 			// A script node has either zero or one child. Can't be more because
@@ -1207,11 +1267,22 @@ HtmlProcessor::processJavaScript(HtmlNode* script, bool& nuke_adjacent_noscript)
 			processDocFragmentNode(script);
 			return;
 		}
-		auto_ptr<ErrorDescriptor> edesc = fetchScript(
+	
+		
+		if (m_ptrRequestMetadata->headers().hasHeader(BString("Authorization"))) {
+			// Fetching a script would probably fail, so we don't even try.
+			if (m_pScriptContext) {
+				m_pScriptContext->discardSubprocessorOutput();
+			}
+			processDocFragmentNode(script);
+			return;
+		}
+		
+		fetch_res = fetchScript(
 			prepareScriptFetchRequest(script_url), script_code,
-			max_body_size, max_fetch_size, redirects, cookies
+			max_body_size, max_fetch_size
 		);
-		if (edesc.get()) {
+		if (fetch_res.errorDescriptor.get()) {
 			//DEBUGLOG("Script fetching failed: " << edesc->getErrorMessage());
 			if (m_pScriptContext) {
 				m_pScriptContext->discardSubprocessorOutput();
@@ -1235,10 +1306,10 @@ HtmlProcessor::processJavaScript(HtmlNode* script, bool& nuke_adjacent_noscript)
 	}
 	if (script_external) {
 		context->setExternalScriptInvolved();
-		if (redirects) {
+		if (fetch_res.numRedirects != 0) {
 			context->setRedirectInvolved();
 		}
-		if (cookies) {
+		if (fetch_res.persistentCookies) {
 			context->setPersistentCookieInvolved();
 		}
 	}
@@ -1252,8 +1323,8 @@ HtmlProcessor::processJavaScript(HtmlNode* script, bool& nuke_adjacent_noscript)
 	
 	if (!m_pScriptContext) {
 		auto_ptr<AbstractScriptOperation> operation = processJavaScriptOutput(
-			*context, script, script_code,
-			script_src_str, nuke_adjacent_noscript
+			*context, script, script_code, script_src_str,
+			fetch_res.dynamicContent, nuke_adjacent_noscript
 		);
 		operation->performOperation();
 	} else {
@@ -1268,7 +1339,7 @@ HtmlProcessor::executeJavaScript(
 	ScriptContext& context, HtmlNode* script,
 	SplittableBuffer const& script_code, string const& script_src)
 {
-	SubProcessor subprocessor(*this, context);
+	SubProcessor subprocessor(*this, context, SubProcessor::DOCUMENT_WRITE);
 	
 	static char const js_str[] = { 'j','a','v','a','s','c','r','i','p','t' };
 	BString language = script->getAttribute(BString("language"));
@@ -1279,15 +1350,15 @@ HtmlProcessor::executeJavaScript(
 		js_version = language.toStdString();
 	}
 	
-	m_jsEnv->executeScript(script_code.toString(), script_src, js_version.c_str());
+	m_jsEnv->executeScript(script_code.toBString(), script_src.c_str(), js_version.c_str());
 	context.finishScriptOutput();
 }
 
 auto_ptr<HtmlProcessor::AbstractScriptOperation>
 HtmlProcessor::processJavaScriptOutput(
 	ScriptContext& context, HtmlNode* script,
-	SplittableBuffer const& script_code,
-	string const& script_src, bool& nuke_adjacent_noscript)
+	SplittableBuffer const& script_code, string const& script_src,
+	bool dynamic_content, bool& nuke_adjacent_noscript)
 {
 	auto_ptr<AbstractScriptOperation> res;
 	
@@ -1334,79 +1405,168 @@ HtmlProcessor::processJavaScriptOutput(
 		//	<< ", location: " << suspect->getLocation());
 		
 		HeuristicScore::Status const status = suspect->getScore().getStatus();
-		switch (status) {
-			case HeuristicScore::AD: {
+		
+		if (status == HeuristicScore::PROBABLY_AD ||
+		    status == HeuristicScore::PROBABLY_NOT_AD) {
+			res = tryMarkScriptForAnalyzing(
+				context, *suspect, script, script_code, script_src
+			);
+			if (res.get()) {
+				break;
+			}
+		}
+		
+		if (status == HeuristicScore::AD ||
+		    status == HeuristicScore::PROBABLY_AD) {
+			if (status == HeuristicScore::AD) {
 				script->getParent()->setAdMark(true);
-				if (!context.isCssInvolved() ||
-				    context.isSubprocessorOutputDiscarded()) {
-					res = createScriptSubstitution(script, *suspect);
-				} else {
-					BString const prefix(
-						"<!--\r\n[BFilter] Script has been "
-						"replaced with its (modified) output: \r\n-->"
-					);
-					BString const suffix(
-						"<!--\r\n[BFilter] Script output end.\r\n-->"
-					);
-					SplittableBuffer data;
-					data.appendDestructive(context.subprocessorOutput());
-					data.prepend(prefix);
-					data.append(suffix);
-					res.reset(new RemoveScriptOperation(
-						*this, script, data.toBString()
-					));
-					nuke_adjacent_noscript = true;
-				}
-				break;
 			}
-			case HeuristicScore::PROBABLY_AD:
-			case HeuristicScore::PROBABLY_NOT_AD: {
-				int num_marks = 0;
-				SplittableBuffer new_body;
-				if (suspect->isLocationAbsolute() &&
-				    InsensitiveEqual()(suspect->getLocation().getScheme(), BString("http"))) {
-					num_marks = context.adSuspects().markForAnalyzing(
-						script_code, new_body, HeuristicScore::PROBABLY_NOT_AD
-					);
-				}
-				if (num_marks != 0) {
-					BString comment;
-					if (script_src.empty()) {
-						comment = BString(
-							"<!--\r\n[BFilter] The following script has been modified.\r\n-->"
-						);
-					} else {
-						BString const prefix(
-							"<!--\r\n[BFilter] The following script has been fetched from \""
-						);
-						BString const suffix("\" and modified.\r\n-->");
-						SBOutStream strm(prefix.size() + script_src.size() + suffix.size());
-						strm << prefix << script_src << suffix;
-						comment = strm.data().toBString();
-					}
-					res.reset(new ReplaceBodyScriptOperation(*this, script, new_body, comment));
-				} else {
-					// impossible to mark for analyzing
-					if (status == HeuristicScore::PROBABLY_AD) {
-						res = createScriptSubstitution(script, *suspect);
-					}
-				}
-				break;
-			}
-			default: break;
+			res = substituteAdScript(
+				context, *suspect, script,
+				script_code, script_src, nuke_adjacent_noscript
+			);
 		}
 	} while (false);
 	
 	if (!res.get()) {
+		/*
+		In case the script is external, we may do one of the following:
+		1. Inline the script, since we have already fetched it.
+		2. Leave it as is.
+		
+		We have to consider the following points:
+		a) Inlining a script could improve performance, especially if
+		   it's not cacheable.
+		b) Some scripts look for themselves in the DOM (using the "src" property)
+		   to find the relative path to itself and to load other scripts from the
+		   same directory. In this case, inlinig a script would prevent it from
+		   working properly.
+		c) An advertising engine may choose not to generate an ad under some conditions.
+		   For example, a certain engine produces empty (actually not empty but no-op)
+		   scripts from time to time.  Not inlining such a script would probably mean
+		   a missed ad.
+		
+		Having considered the above points, I came up with the following:
+		
+		if (script is external) {
+			if (script is dynamically generated) {
+				inline it;
+			} else {
+				leave it as is;
+			}
+		}
+		
+		Currently we determine whether the script is dynamically generated
+		by the absence of Last-Modified response header.
+		*/
+		
 		BString comment;
-		if (!script_src.empty()) {
+		SplittableBuffer const* inline_code = 0;
+		
+		if (!script_src.empty() && dynamic_content) {
+			inline_code = &script_code;
 			BString const prefix("<!--\r\n[BFilter] The following script has been fetched from \"");
 			BString const suffix("\"\r\n-->");
-			SBOutStream strm(prefix.size() + script_src.size() + suffix.size());
-			strm << prefix << script_src << suffix;
-			comment = strm.data().toBString();
+			
+			TempStreamScope scope(m_tempStream);
+			m_tempStream << prefix << script_src << suffix;
+			comment = m_tempStream.data().toBString();
 		}
-		res.reset(new LeaveAsIsScriptOperation(*this, script, script_code, comment));
+		res.reset(new LeaveAsIsScriptOperation(*this, script, inline_code, comment));
+	}
+	
+	return res;
+}
+
+/*
+Marking a script for analyzing is a matter of replacing the intances
+of ad host (which we already know) with ad_host/bf-analyze
+*/
+auto_ptr<HtmlProcessor::AbstractScriptOperation>
+HtmlProcessor::tryMarkScriptForAnalyzing(
+	ScriptContext& context, AdSuspect const& suspect, HtmlNode* script,
+	SplittableBuffer const& script_code, string const& script_src)
+{
+	auto_ptr<AbstractScriptOperation> res;
+	
+	if (context.isInnerHtmlInvolved()) {
+		return res;
+	}
+	
+	if (!context.adSuspects().isSameHost()) {
+		// Better avoid such complex cases,
+		// although theoretically it might work.
+		return res;
+	}
+	
+	int num_marks = 0;
+	SplittableBuffer new_body;
+	if (suspect.isLocationAbsolute() &&
+		InsensitiveEqual()(suspect.getLocation().getScheme(), BString("http"))) {
+		num_marks = context.adSuspects().markForAnalyzing(
+			script_code, new_body, HeuristicScore::PROBABLY_NOT_AD
+		);
+	}
+	if (num_marks == 0) {
+		return res;
+	}
+	
+	BString comment;
+	if (script_src.empty()) {
+		comment = BString(
+			"<!--\r\n[BFilter] The following script has been modified.\r\n-->"
+		);
+	} else {
+		BString const prefix(
+			"<!--\r\n[BFilter] The following script has been fetched from \""
+		);
+		BString const suffix("\" and modified.\r\n-->");
+		
+		TempStreamScope scope(m_tempStream);
+		m_tempStream << prefix << script_src << suffix;
+		comment = m_tempStream.data().toBString();
+	}
+	
+	res.reset(new ReplaceBodyScriptOperation(*this, script, new_body, comment));
+	return res;
+}
+
+auto_ptr<HtmlProcessor::AbstractScriptOperation>
+HtmlProcessor::substituteAdScript(
+	ScriptContext& context, AdSuspect const& suspect, HtmlNode* script,
+	SplittableBuffer const& script_code, string const& script_src,
+	bool& nuke_adjacent_noscript)
+{
+	auto_ptr<AbstractScriptOperation> res;
+	
+	if (context.isCssInvolved() &&
+	    !context.isInnerHtmlInvolved() &&
+	    !context.isSubprocessorOutputDiscarded()) {
+		/*
+		Consider a case where a script outputs the following:
+		<div id="x"><a ...><img ...></a></div>
+		Now imagine that there is a css rule that makes
+		element x absolute positioned. In this case it's better
+		to keep the output structure, or our ad replacement will
+		appear in a wrong place, possibly covering other content.
+		*/
+		BString const prefix(
+			"<!--\r\n[BFilter] Script has been "
+			"replaced with its (modified) output: \r\n-->"
+		);
+		BString const suffix(
+			"<!--\r\n[BFilter] Script output end.\r\n-->"
+		);
+		SplittableBuffer data;
+		data.appendDestructive(context.subprocessorOutput());
+		data.prepend(prefix);
+		data.append(suffix);
+		res.reset(new RemoveScriptOperation(
+			*this, script, data.toBString()
+		));
+		nuke_adjacent_noscript = true;
+	} else {
+		res = createScriptSubstitution(script, suspect);
 	}
 	
 	return res;
@@ -1427,6 +1587,7 @@ HtmlProcessor::processImg(HtmlNode* img, bool explicit_empty, HtmlNode const* an
 		WebbugHeuristicScore score;
 		score.processPageHint(m_pageHint);
 		score.processImageUrl(imgurl, imgurl_rel, m_baseURL);
+		score.processNoscript(m_isInsideNoscript);
 		if (m_pScriptContext) {
 			score.processScriptLevel(
 				m_pScriptContext->getNestLevel(),
@@ -1478,15 +1639,32 @@ HtmlProcessor::processImg(HtmlNode* img, bool explicit_empty, HtmlNode const* an
 			m_pScriptContext->incUnclickableImageCount();
 		}
 	} else {
-		URI aurl_rel(anchor->getAttribute(BString("href")));
+		BString anchor_link;
+		BString anchor_target;
+		
+		BString onclick = anchor->getAttribute(BString("onclick"));
+		if (!onclick.empty()) {
+			PageOpenListener listener(*m_jsEnv);
+			m_jsEnv->executeScriptAsFunction(onclick, "<onclick>", 1);
+			anchor_link = listener.getURL();
+			anchor_target = listener.getTarget();
+		}
+		
+		if (anchor_link.empty()) {
+			anchor_link = anchor->getAttribute(BString("href"));
+			anchor_target = anchor->getAttribute(BString("target"));
+		}
+		
+		URI aurl_rel(anchor_link);
 		URI aurl(m_baseURL, aurl_rel);
 		
 		ImageHeuristicScore score;
 		score.processPageHint(m_pageHint);
 		score.processSize(width, height);
 		score.processUrls(m_baseURL, imgurl, imgurl_rel, aurl, aurl_rel);
-		score.processLinkTarget(anchor->getAttribute(BString("target")));
+		score.processLinkTarget(anchor_target);
 		score.processAssociatedMap(img->hasAttribute(BString("usemap")));
+		score.processNoscript(m_isInsideNoscript);
 		//DEBUGLOG("imgurl = " << imgurl);
 		//DEBUGLOG("score = " << score.getNumericScore());
 		if (m_pScriptContext) {
@@ -1552,12 +1730,10 @@ HtmlProcessor::processOpeningTagAsText(HtmlNode* node, bool explicit_empty)
 	assert(!node->getParent());
 	
 	if (m_docFragment.hasChildren()) {
-		SBOutStream strm(100);
-		openingTagToStream(strm, node, m_contentType, explicit_empty);
-		tagTextAppend(
-			m_pContainingNode,
-			strm.data().begin(), strm.data().end()
-		);
+		TempStreamScope scope(m_tempStream);
+		openingTagToStream(m_tempStream, node, m_contentType, explicit_empty);
+		SplittableBuffer const& data = m_tempStream.data();
+		tagTextAppend(m_pContainingNode, data.begin(), data.end());
 	} else {
 		openingTagToStream(m_processedData, node, m_contentType, explicit_empty);
 	}
@@ -1598,7 +1774,7 @@ HtmlProcessor::processDocFragmentNodeReplacement(HtmlNode* original, HtmlNode* r
 	if (isAncestorOrSelf(original, m_pContainingNode)) {
 		m_pContainingNode = original->getParent();
 	}
-	replacement->insertBefore(original);
+	original->insertPrevSibling(replacement);
 	original->detachFromParent();
 	processNode(replacement, true, m_docFragment.getFirstChild() == replacement);
 	outputDocFragmentLeadingText();
@@ -1643,20 +1819,19 @@ HtmlProcessor::processNode(HtmlNode* node, bool skip_stoppers, bool can_output_d
 		if (can_output_directly) {
 			openingTagToStream(m_processedData, node, m_contentType, explicit_empty);
 		} else {
-			SBOutStream strm(100);
-			openingTagToStream(strm, node, m_contentType, explicit_empty);
+			TempStreamScope scope(m_tempStream);
+			openingTagToStream(m_tempStream, node, m_contentType, explicit_empty);
 			HtmlNodePtr text_node(HtmlTextNode::create());
-			text_node->appendContent(
-				strm.data().begin(), strm.data().end()
-			);
-			text_node->insertBefore(node);
+			SplittableBuffer const& data = m_tempStream.data();
+			text_node->appendContent(data.begin(), data.end());
+			node->insertPrevSibling(text_node);
 		}
 	}
 	
 	for (HtmlNodePtr child(node->getFirstChild()); child; ) {
 		HtmlNode* next_child = child->getNextSibling();
 		child->detachFromParent();
-		child->insertBefore(node);
+		node->insertPrevSibling(child);
 		can_output_directly = processNode(child.get(), skip_stoppers, can_output_directly, false);
 		child.reset(next_child);
 	}
@@ -1673,7 +1848,7 @@ HtmlProcessor::processNode(HtmlNode* node, bool skip_stoppers, bool can_output_d
 		} else {
 			HtmlNodePtr text_node(HtmlTextNode::create());
 			text_node->appendContent(BString(ctag));
-			text_node->insertBefore(node);
+			node->insertPrevSibling(text_node);
 		}
 	}
 	
@@ -1908,25 +2083,22 @@ HtmlProcessor::tagTerminatesAnchor(Iterator const& begin, Iterator const& end)
 	);
 }
 
-auto_ptr<ErrorDescriptor>
+HtmlProcessor::ScriptFetchResult
 HtmlProcessor::fetchScript(
 	RequestPtr const& request, SplittableBuffer& output,
-	size_t max_body_size, size_t max_fetch_size,
-	bool& redirects, bool& cookies)
+	size_t max_body_size, size_t max_fetch_size)
 {
-	auto_ptr<ErrorDescriptor> edesc;
-	int num_redirects = 0;
-	cookies = false;
+	ScriptFetchResult res;
 	
 	{
 		URI const& uri = request->requestLine().getURI();
 		GlobalState::ReadAccessor global_state;
 		if (global_state->urlPatterns().getSubstitutionFor(uri)) {
-			return edesc;
+			return res; // as if we fetched it but it was empty
 		}
-		if (global_state->urlPatterns().isForbidden(uri)) {
-			edesc.reset(new ErrorDescriptor(int(ErrorCodes::URL_FORBIDDEN)));
-			return edesc;
+		if (global_state->urlPatterns().isFilteringDisabled(uri)) {
+			res.errorDescriptor.reset(new ErrorDescriptor(int(ErrorCodes::SCRIPT_FETCH_NOFILTER)));
+			return res;
 		}
 	}
 	
@@ -1944,8 +2116,9 @@ HtmlProcessor::fetchScript(
 			
 			if (HttpResponseMetadata const* metadata = fetcher.responseMetadata().get()) {
 				if (metadata->statusLine().isRedirect()) {
-					if (++num_redirects > MAX_SCRIPT_FETCH_REDIRECTS) {
-						edesc.reset(new ErrorDescriptor(
+					res.dynamicContent = true;
+					if (++res.numRedirects > MAX_SCRIPT_FETCH_REDIRECTS) {
+						res.errorDescriptor.reset(new ErrorDescriptor(
 							ErrorCodes::TOO_MANY_REDIRECTS,
 							"too many redirects"
 						));
@@ -1963,14 +2136,15 @@ HtmlProcessor::fetchScript(
 		}
 		
 		if (HttpResponseMetadata const* metadata = fetcher.responseMetadata().get()) {
-			cookies |= hasPersistentCookies(*metadata);
+			res.persistentCookies = (res.persistentCookies || hasPersistentCookies(*metadata));
+			res.dynamicContent = (res.dynamicContent || isDynamicResource(*metadata));
 		}
 		
 		if (follow_redirect) {
 			continue;
 		}
 		
-		if (!edesc.get()) {
+		if (!res.errorDescriptor.get()) {
 			if (fetcher.getStatus() == HttpFetcher::COMPLETE) {
 				int const status = fetcher.responseMetadata()->statusLine().getCode();
 				if (status < 200 || status >= 400) {
@@ -1978,17 +2152,17 @@ HtmlProcessor::fetchScript(
 					fetcher.responseMetadata()->statusLine().toStream(strm);
 					string status(strm.str());
 					status.resize(status.size() - 2); // remove the training \r\n
-					edesc.reset(new ErrorDescriptor(
+					res.errorDescriptor.reset(new ErrorDescriptor(
 						ErrorCodes::WRONG_RESPONSE_STATUS, status
 					));
 				} else {
 					output.appendDestructive(fetcher.responseBody());
 				}
 			} else if (fetcher.getStatus() == HttpFetcher::FAILED) {
-				edesc = fetcher.errorDescriptor();
+				res.errorDescriptor = fetcher.errorDescriptor();
 			} else {
 				// Reactor::handle_events() returned -1?
-				edesc.reset(new ErrorDescriptor(
+				res.errorDescriptor.reset(new ErrorDescriptor(
 					ErrorCodes::UNKNOWN_ERROR,
 					"unknown error"
 				));
@@ -1997,8 +2171,7 @@ HtmlProcessor::fetchScript(
 		break;
 	}
 	
-	redirects = num_redirects;
-	return edesc;
+	return res;
 }
 
 RequestPtr
@@ -2037,8 +2210,9 @@ HtmlProcessor::prepareScriptFetchRequest(URI const& url)
 			request->headers().setHeader(*hdr);
 		}
 	}
-	request->headers().setHeader(HttpHeader(BString("Accept"), BString("*" "/" "*")));
-	request->headers().setHeader(HttpHeader(BString("Referer"), BString(m_baseURL.toString())));
+	request->headers().setHeader(BString("Accept"), BString("*" "/" "*"));
+	request->headers().setHeader(BString("Accept-Language"), BString("en-us,en;q=0.5"));
+	request->headers().setHeader(BString("Referer"), BString(m_baseURL.toString()));
 	request->setBodyStatus(HttpRequestMetadata::BODY_FORBIDDEN);
 	return request;
 }
@@ -2097,6 +2271,12 @@ HtmlProcessor::isPersistentCookie(std::string const& str)
 	return false;
 }
 
+bool
+HtmlProcessor::isDynamicResource(HttpResponseMetadata const& metadata)
+{
+	return !metadata.headers().hasHeader(BString("Last-Modified"));
+}
+
 auto_ptr<HtmlProcessor::AbstractScriptOperation>
 HtmlProcessor::createScriptSubstitution(HtmlNode* node, AdSuspect const& ad)
 {
@@ -2114,6 +2294,7 @@ HtmlProcessor::createScriptSubstitution(HtmlNode* node, AdSuspect const& ad)
 
 HtmlProcessor::ScriptContext::ScriptContext(JsEnvironment& env)
 :	m_nestLevel(0),
+	m_innerHtmlLevel(0),
 	m_flags(0),
 	m_descendantScriptFetches(0),
 	m_unclickableImageCount(0),
@@ -2123,10 +2304,9 @@ HtmlProcessor::ScriptContext::ScriptContext(JsEnvironment& env)
 	m_scriptOutputSize(0),
 	m_webbugTailEvidence(0),
 	m_pSubProcessor(0),
-	m_subprocessorOutputSize(0),
-	m_listenerHandle(this)
+	m_subprocessorOutputSize(0)
 {
-	env.setListener(m_listenerHandle);
+	env.setListener(*this);
 }
 
 HtmlProcessor::ScriptContext::~ScriptContext()
@@ -2134,16 +2314,31 @@ HtmlProcessor::ScriptContext::~ScriptContext()
 }
 
 void
-HtmlProcessor::ScriptContext::processJsOutput(char const* addr, size_t len)
+HtmlProcessor::ScriptContext::processJsOutput(char const* data, bool newline)
 {
-	if (len == 3 && m_scriptOutputSize == 0 && strncmp(addr, "-->", 3)) {
+	if (m_innerHtmlLevel != 0) {
+		// Scripts inserted using innerHTML can't use document.write().
+		return;
+	}
+	
+	if (m_scriptOutputSize == 0 && strncmp(data, "-->", 3) == 0) {
 		m_webbugTailEvidence |= WEBBUG_TAIL_EV1;
 	}
+	
+	size_t len = strlen(data);
 	m_scriptOutputSize += len;
 	if (m_pSubProcessor) {
-		SplittableBuffer data;
-		data.append(BString(addr, len));
-		m_pSubProcessor->consume(data, false);
+		SplittableBuffer buf;
+		{
+			SBOutStream& temp_stream = m_pSubProcessor->tempStream();
+			TempStreamScope scope(temp_stream);
+			temp_stream.write(data, len);
+			if (newline) {
+				temp_stream << "\r\n";
+			}
+			temp_stream.swapData(buf);
+		}
+		m_pSubProcessor->consume(buf, false);
 		consumeSubprocessorOutput();
 	}
 }
@@ -2154,6 +2349,28 @@ HtmlProcessor::ScriptContext::processOnLoadAssignment()
 	setOnLoadAssigned();
 }
 
+void
+HtmlProcessor::ScriptContext::processInnerHtmlAssignment(char const* data)
+{
+	discardSubprocessorOutput();
+	
+	if (!m_pSubProcessor) {
+		// should not happen
+		return;
+	}
+	
+	SplittableBuffer buf;
+	{
+		SBOutStream& temp_stream = m_pSubProcessor->tempStream();
+		TempStreamScope scope(temp_stream);
+		temp_stream << data;
+		temp_stream.swapData(buf);
+	}
+	
+	SubProcessor subprocessor(*m_pSubProcessor, *this, SubProcessor::INNER_HTML);
+	subprocessor.consume(buf, true);
+}
+
 void
 HtmlProcessor::ScriptContext::finishScriptOutput()
 {
@@ -2197,27 +2414,56 @@ HtmlProcessor::ScriptContext::processNestedScriptUrl(URI const& url, URI const&
 {
 	typedef HeuristicScore HS;
 	if (HS::getUrlRelationship(url, page_url) == HS::URLS_UNRELATED) {
-		deque<URI>::iterator it(m_unrelatedNestedScripts.begin());
-		deque<URI>::iterator const end(m_unrelatedNestedScripts.end());
+		vector<URI>::iterator it(m_unrelatedNestedScripts.begin());
+		vector<URI>::iterator const end(m_unrelatedNestedScripts.end());
 		for (; it != end; ++it) {
 			if (HS::getUrlRelationship(url, *it) == HS::URLS_UNRELATED) {
 				m_flags |= TOO_MANY_UNRELATED_SCRIPTS;
 				break;
 			}
 		}
+		
+		m_unrelatedNestedScripts.reserve(5);
 		m_unrelatedNestedScripts.push_back(url);
 	}
 }
 
 
+/*====================== HtmlProcessor::SubProcessor ======================*/
+
+HtmlProcessor::SubProcessor::SubProcessor(
+	HtmlProcessor& parent, ScriptContext& context, Type type)
+:	HtmlProcessor(parent, context),
+	m_rContext(context),
+	m_oldNestLevel(context.getNestLevel()),
+	m_oldInnerHtmlLevel(context.getInnerHtmlLevel()),
+	m_pOldSubProcessor(context.getSubProcessor())
+{
+	m_rContext.setSubProcessor(this);
+	m_rContext.setNestLevel(m_oldNestLevel + 1);
+	if (type == INNER_HTML) {
+		m_rContext.setInnerHtmlInvolved();
+		m_rContext.setInnerHtmlLevel(m_oldInnerHtmlLevel + 1);
+	}
+	
+}
+
+HtmlProcessor::SubProcessor::~SubProcessor()
+{
+	m_rContext.setInnerHtmlLevel(m_oldInnerHtmlLevel);
+	m_rContext.setNestLevel(m_oldNestLevel);
+	m_rContext.setSubProcessor(m_pOldSubProcessor);
+}
+
+
 /*=============== HtmlProcessor::LeaveAsIsScriptOperation =================*/
 
 HtmlProcessor::LeaveAsIsScriptOperation::LeaveAsIsScriptOperation(
 	HtmlProcessor& processor, HtmlNode* node,
-	SplittableBuffer const& code, BString const& comment)
+	SplittableBuffer const* code, BString const& comment)
 :	m_rProcessor(processor),
 	m_pNode(node),
-	m_rCode(code),
+	m_pCode(code),
 	m_comment(comment)
 {
 }
@@ -2229,12 +2475,12 @@ HtmlProcessor::LeaveAsIsScriptOperation::~LeaveAsIsScriptOperation()
 void
 HtmlProcessor::LeaveAsIsScriptOperation::performOperation()
 {
-	if (m_pNode->hasAttribute(BString("src"))) {
-		m_rProcessor.inlineExternalScript(m_pNode, m_rCode);
+	if (m_pCode && m_pNode->hasAttribute(BString("src"))) {
+		m_rProcessor.inlineExternalScript(m_pNode, *m_pCode);
 	}
 	HtmlNodePtr comment_node(HtmlTextNode::create());
 	comment_node->appendContent(m_comment);
-	comment_node->insertBefore(m_pNode);
+	m_pNode->insertPrevSibling(comment_node);
 	m_rProcessor.processDocFragmentNode(m_pNode);
 }
 
@@ -2286,6 +2532,31 @@ HtmlProcessor::ReplaceBodyScriptOperation::performOperation()
 	m_pNode->appendChild(content_node.get());
 	HtmlNodePtr comment_node(HtmlTextNode::create());
 	comment_node->appendContent(m_comment);
-	comment_node->insertBefore(m_pNode);
+	m_pNode->insertPrevSibling(comment_node);
 	m_rProcessor.processDocFragmentNode(m_pNode);
 }
+
+
+/*================== HtmlProcessor::PageOpenListener ===================*/
+
+HtmlProcessor::PageOpenListener::PageOpenListener(JsEnvironment& js_env)
+{
+	js_env.setPageOpenListener(*this);
+}
+
+HtmlProcessor::PageOpenListener::~PageOpenListener()
+{
+}
+
+void
+HtmlProcessor::PageOpenListener::processPageOpening(
+	char const* url, char const* target)
+{
+	assert(url);
+	assert(target);
+	
+	if (m_url.empty()) {
+		m_url = BString(url, strlen(url));
+		m_target = BString(target, strlen(target));
+	}
+}
