@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -50,15 +50,17 @@
 #include "HttpMessageShaper.h"
 #include "GlobalState.h"
 #include "ServiceContext.h"
-#include "InetAddr.h"
+#include "SymbolicInetAddr.h"
 #include "ProxyDescriptor.h"
+#include "DownloadProgress.h"
 #include "URI.h"
+#include "TimeDelta.h"
 #include "StringUtils.h"
 #include "InsensitiveEqual.h"
 #include "NetworkActivityReporter.h"
+#include "RequestLog.h"
+#include "RequestTag.h"
 #include "Debug.h"
-#include <ace/Time_Value.h>
-#include <ace/INET_Addr.h>
 #include <algorithm>
 #include <list>
 #include <cassert>
@@ -74,6 +76,7 @@ class Server::Request :
 public:
 	Request(ServiceContext& context, ServerWeakPtr const& server_ptr,
 		ConstRequestPtr const& request_metadata,
+		RequestTag const& request_tag,
 		IntrusivePtr<AbstractResponseHandler> const& response_handler,
 		std::auto_ptr<FilterTryList> filter_try_list);
 	
@@ -99,9 +102,9 @@ public:
 	
 	bool restart();
 	
-	void setErrorResponse(std::auto_ptr<ErrorDescriptor> edesc) { m_ptrErrDesc = edesc; }
+	void setErrorResponse(IntrusivePtr<ErrorDescriptor> const& edesc) { m_ptrErrDesc = edesc; }
 	
-	std::auto_ptr<ErrorDescriptor>& errorDescriptor() { return m_ptrErrDesc; }
+	IntrusivePtr<ErrorDescriptor> const& errorDescriptor() { return m_ptrErrDesc; }
 	
 	std::string const& getErrorMsg() const { return m_errorMsg; }
 	
@@ -117,15 +120,19 @@ public:
 	
 	HttpRequestMetadata const& getOutgoingMetadata() const { return *m_ptrOutgoingMetadata; }
 	
-	InetAddr const& getNextHopAddr() const { return m_nextHopAddr; }
+	RequestTag const& getTag() const { return m_tag; }
+	
+	SymbolicInetAddr const& getNextHopAddr() const { return m_nextHopAddr; }
+	
+	void updateRoute();
 protected:
-	virtual void raiseError(std::auto_ptr<ErrorDescriptor> edesc);
+	virtual void raiseError(IntrusivePtr<ErrorDescriptor> const& edesc);
 	
 	virtual bool isErrorRaised() const;
 private:
 	virtual void appendBodyData(SplittableBuffer& data, bool eof);
 	
-	virtual void cancelRequest(std::auto_ptr<ErrorDescriptor> edesc);
+	virtual void cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc);
 	
 	virtual void processNewData(SplittableBuffer& data, bool eof);
 	
@@ -137,10 +144,11 @@ private:
 	ServerWeakPtr m_ptrServer;
 	ConstRequestPtr m_ptrOrigMetadata;
 	std::auto_ptr<HttpRequestMetadata> m_ptrOutgoingMetadata;
+	RequestTag m_tag;
 	std::auto_ptr<ResponseFilterChain> m_ptrFilterChain;
-	InetAddr m_nextHopAddr;
+	SymbolicInetAddr m_nextHopAddr;
 	bool m_willGoThroughProxy;
-	std::auto_ptr<ErrorDescriptor> m_ptrErrDesc;
+	IntrusivePtr<ErrorDescriptor> m_ptrErrDesc;
 	std::string m_errorMsg;
 	SplittableBuffer m_data;
 	bool m_isEOF;
@@ -149,12 +157,13 @@ private:
 };
 
 
-class Server::PhantomRequest :
-	public AbstractRequestHandler,
-	private NonCopyable
+class Server::PhantomRequest : public AbstractRequestHandler
 {
+	DECLARE_NON_COPYABLE(PhantomRequest)
 public:
-	PhantomRequest(ConstRequestPtr const& response_metadata,
+	PhantomRequest(
+		ConstRequestPtr const& response_metadata,
+		RequestTag const& request_tag,
 		IntrusivePtr<AbstractResponseHandler> const& response_handler,
 		CraftedResponseGenerator const& response_generator);
 	
@@ -167,12 +176,17 @@ public:
 	URI const& getRequestURI() const { return m_ptrOrigMetadata->requestLine().getURI(); }
 	
 	int getClientRequestId() const { return m_ptrOrigMetadata->getClientRequestId(); }
+	
+	bool isHeadRequest() const { return m_ptrOrigMetadata->requestLine().getMethod() == BString("HEAD"); }
+	
+	RequestTag const& getTag() const { return m_tag; }
 private:
 	virtual void appendBodyData(SplittableBuffer& data, bool eof);
 	
-	virtual void cancelRequest(std::auto_ptr<ErrorDescriptor> edesc);
+	virtual void cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc);
 	
 	ConstRequestPtr m_ptrOrigMetadata;
+	RequestTag m_tag;
 	IntrusivePtr<AbstractResponseHandler> m_ptrResponseHandler;
 	CraftedResponseGenerator m_responseGenerator;
 };
@@ -193,25 +207,41 @@ private:
 };
 
 
+class Server::RequestCanceller
+{
+public:
+	void operator()(IntrusivePtr<Request> const& req) {
+		RequestLog::logRequestCancel(req->getTag());
+	}
+	
+	void operator()(IntrusivePtr<PhantomRequest> const& req) {
+		if (req) {
+			RequestLog::logRequestCancel(req->getTag());
+		}
+	}
+};
+
+
 Server::Server(ServiceContext& context, ServerTimeouts const& timeouts)
 :	m_rContext(context),
 	m_timeouts(timeouts),
 	m_pipelineSize(0),
+	m_connectingTo(string(), -1),
 	m_lastResponseVersion(0, 0),
 	m_readTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Server::onReadTimeout),
-		ACE_Time_Value(m_timeouts.getReadTimeout())
+		TimeDelta::fromSec(m_timeouts.getReadTimeout()),
+		&context.reactor()
 	),
 	m_writeTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Server::onWriteTimeout),
-		ACE_Time_Value(m_timeouts.getWriteTimeout())
+		TimeDelta::fromSec(m_timeouts.getWriteTimeout()),
+		&context.reactor()
 	),
 	m_delayedActionTimer(
-		context.reactor(),
 		sigc::mem_fun(*this, &Server::onDelayedAction),
-		ACE_Time_Value::zero
+		TimeDelta::zero(),
+		&context.reactor()
 	),
 	m_eventMask(Reactor::NO_EVENTS),
 	m_flags(0),
@@ -219,25 +249,28 @@ Server::Server(ServiceContext& context, ServerTimeouts const& timeouts)
 	m_ptrSelf(this)
 {
 	m_rContext.reactor().beforeSleepSignal().connect(sigc::mem_fun(*this, &Server::updateState));
+	m_connector.setListener(this);
 }
 
 Server::~Server()
 {
+	m_connector.removeListener();
 	dropConnection();
 }
 
 IntrusivePtr<AbstractRequestHandler>
 Server::submitRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler)
 {
 	std::auto_ptr<FilterTryList> filter_try_list(new FilterTryList);
 	IntrusivePtr<Request> request(new Request(
-		m_rContext, m_ptrSelf, request_metadata, handler, filter_try_list
+		m_rContext, m_ptrSelf, request_metadata,
+		request_tag, handler, filter_try_list
 	));
 	inspectRequest(*request);
-	m_requestQueue.push_back(request);
-	m_phantomRequestQueue.push_back(IntrusivePtr<PhantomRequest>());
+	enqueueRequest(request);
 	updatePipeline();
 	return request;
 }
@@ -245,15 +278,16 @@ Server::submitRequest(
 IntrusivePtr<AbstractRequestHandler>
 Server::submitRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler,
 	std::auto_ptr<FilterTryList> filter_try_list)
 {
 	IntrusivePtr<Request> request(new Request(
-		m_rContext, m_ptrSelf, request_metadata, handler, filter_try_list
+		m_rContext, m_ptrSelf, request_metadata,
+		request_tag, handler, filter_try_list
 	));
 	inspectRequest(*request);
-	m_requestQueue.push_back(request);
-	m_phantomRequestQueue.push_back(IntrusivePtr<PhantomRequest>());
+	enqueueRequest(request);
 	updatePipeline();
 	return request;
 }
@@ -261,29 +295,32 @@ Server::submitRequest(
 IntrusivePtr<AbstractRequestHandler>
 Server::submitRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler,
-	auto_ptr<ErrorDescriptor> error_response)
+	IntrusivePtr<ErrorDescriptor> const& error_response)
 {
 	std::auto_ptr<FilterTryList> filter_try_list(new FilterTryList);
 	IntrusivePtr<Request> request(new Request(
-		m_rContext, m_ptrSelf, request_metadata, handler, filter_try_list
+		m_rContext, m_ptrSelf, request_metadata,
+		request_tag, handler, filter_try_list
 	));
 	request->setErrorResponse(error_response);
-	m_requestQueue.push_back(request);
-	m_phantomRequestQueue.push_back(IntrusivePtr<PhantomRequest>());
+	enqueueRequest(request);
 	return request;
 }
 
 IntrusivePtr<AbstractRequestHandler>
 Server::submitRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& handler,
 	CraftedResponseGenerator const& response_generator)
 {
 	IntrusivePtr<PhantomRequest> request(new PhantomRequest(
-		request_metadata, handler, response_generator
+		request_metadata, request_tag,
+		handler, response_generator
 	));
-	m_phantomRequestQueue.push_back(request);
+	enqueuePhantomRequest(request);
 	return request;
 }
 
@@ -296,16 +333,48 @@ Server::reset()
 	assert(!m_readTimer.isActive());
 	assert(!m_writeTimer.isActive());
 	
+	RequestCanceller rc;
+	std::for_each(m_requestQueue.begin(), m_requestQueue.end(), rc);
+	std::for_each(m_phantomRequestQueue.begin(), m_phantomRequestQueue.end(), rc);
+	
+	
 	m_requestQueue.clear();
 	m_phantomRequestQueue.clear();
 	m_pipelineSize = 0;
+	m_connectingTo.set(string(), -1);
+	m_connector.abort();
 	m_inputData.clear();
-	m_downloadProgress.reset();
 	m_lastResponseVersion.set(0, 0);
 	m_delayedActionTimer.deactivate();
 	m_flags &= ~WRITE_BUFFER_FULL;
 }
 
+void
+Server::enqueueRequest(IntrusivePtr<Request> const& req)
+{
+	m_requestQueue.push_back(req);
+	m_phantomRequestQueue.push_back(IntrusivePtr<PhantomRequest>());
+	
+	RequestLog::RequestType type = RequestLog::NORMAL_REQUEST;
+	if (&m_rContext.scriptServer() == this) {
+		type = RequestLog::INTERNAL_REQUEST;
+	} else if (req->getTag()->flags().isSet(RequestTag::REQUEST_ANALYZE)) {
+		type = RequestLog::ANALYZE_REQUEST;
+	}
+	RequestLog::logRequest(req->getTag(), req->getRequestURI(), type);
+}
+
+void
+Server::enqueuePhantomRequest(IntrusivePtr<PhantomRequest> const& req)
+{
+	m_phantomRequestQueue.push_back(req);
+	
+	RequestLog::logRequest(
+		req->getTag(), req->getRequestURI(),
+		RequestLog::SUBST_REQUEST
+	);
+}
+
 void
 Server::updatePipeline()
 {	
@@ -337,8 +406,7 @@ Server::updatePipeline()
 		if (m_pipelineSize > 0) {
 			Request const& prev_req = *m_requestQueue[m_pipelineSize - 1];
 			
-			if (!m_ptrConnection.get() ||
-			    m_ptrConnection->getStatus() != AsyncConnection::ESTABLISHED) {
+			if (!m_ptrConnection.get()) {
 				//DEBUGLOG("We don't have a connection yet");
 				break;
 			}
@@ -393,7 +461,7 @@ Server::inspectRequest(Request& req)
 		static char const http[] = {'h','t','t','p'};
 		BString const& scheme = uri.getScheme();
 		if (!StringUtils::ciEqual(scheme.begin(), scheme.end(), http, http + sizeof(http))) {
-			req.setErrorResponse(auto_ptr<ErrorDescriptor>(
+			req.setErrorResponse(IntrusivePtr<ErrorDescriptor>(
 				new ErrorDescriptor(
 					ErrorCodes::UNSUPPORTED_PROTOCOL,
 					"unsupported protocol",
@@ -416,7 +484,7 @@ Server::inspectRequest(Request& req)
 			list<HttpHeaderElement>::iterator const end = structure.elements().end();
 			for (; it != end; ++it) {
 				if (InsensitiveEqual()(it->getName(), BString("100-continue"))) {
-					req.setErrorResponse(auto_ptr<ErrorDescriptor>(
+					req.setErrorResponse(IntrusivePtr<ErrorDescriptor>(
 						new ErrorDescriptor(
 							ErrorCodes::EXPECTATION_FAILED,
 							"expectation failed",
@@ -461,24 +529,15 @@ Server::requestConnectionFor(Request const& req)
 		m_ptrConnection = ServerConnectionPool::instance()
 			->retrieveConnection(req.getNextHopAddr());
 		if (m_ptrConnection.get()) {
-			assert(m_ptrConnection->getStatus() == AsyncConnection::ESTABLISHED);
 			//DEBUGLOG("Reusing an existing server connection");
-			connectionEstablished();
+			processNewConnection();
 			return;
 		}
 	}
 	//DEBUGLOG("Initiating a new server connection");
-	ACE_Time_Value timeout(m_timeouts.getConnectTimeout());
-	m_ptrConnection.reset(new ServerConnection);
-	m_ptrConnection->setListener(this);
-	m_ptrConnection->initiate(m_rContext.reactor(), req.getNextHopAddr(), &timeout);
-}
-
-bool
-Server::isConnectionReady() const
-{
-	return m_ptrConnection.get() &&
-		m_ptrConnection->getStatus() == AsyncConnection::ESTABLISHED;
+	TimeDelta timeout = TimeDelta::fromSec(m_timeouts.getConnectTimeout());
+	m_connectingTo = req.getNextHopAddr();
+	m_connector.initiate(m_rContext.reactor(), m_connectingTo, &timeout);
 }
 
 bool
@@ -516,7 +575,7 @@ Server::onNewOutput()
 void
 Server::processOutputData()
 {
-	while (m_pipelineSize > 0 && isConnectionReady()) {
+	while (m_pipelineSize > 0 && m_ptrConnection.get()) {
 		if (writeRequestData(*m_requestQueue[m_pipelineSize - 1]) == 0) {
 			break;
 		}
@@ -529,7 +588,19 @@ Server::writeRequestData(Request& req)
 	if (!req.hasData()) {
 		return 0;
 	}
-	
+
+#if defined(ACE_WIN32)
+	/*
+	There are certain buggy software firewalls that have problems with
+	multi-fragment writes. We workaround it by coalescing fragments and
+	doing a single write. Fortunately, we only have to do it for requests
+	(where it has almost no effect on performance), as I haven't
+	encountered a firewall that has problems with fragmented responses.
+	*/
+	BString buf(req.data().toBString());
+	ssize_t sent = m_ptrConnection->peer().send(buf.begin(), buf.size());
+	buf.clear(); // just a small optimization of memory usage
+#else
 	iovec iov[ACE_IOV_MAX];
 	int iov_size = 0;
 	for (SplittableBuffer::SpanIterator iter(req.data().spanBegin());
@@ -542,6 +613,7 @@ Server::writeRequestData(Request& req)
 		iov, iov_size
 		// the timeout is infinite, but the socket is in nonblocking mode
 	);
+#endif
 	if (sent < 0 && errno == EWOULDBLOCK) {
 		m_flags |= WRITE_BUFFER_FULL;
 	} else if (sent > 0) {
@@ -620,21 +692,23 @@ Server::processInputData(bool eof)
 		SplittableBuffer body_data;
 		bool body_eof = false;
 		bool const empty_input = m_inputData.empty();
-		bool message_finishes = processNewData(m_inputData, eof, body_data, body_eof);	
-		m_downloadProgress.received() += body_data.size();
+		bool message_finishes = processNewData(m_inputData, eof, body_data, body_eof);
+		req->getTag()->downloadProgress().received() += body_data.size();
 		if (!req->errorDescriptor().get()) {
-#ifdef DEBUG
 			if (message_finishes) {
+				RequestLog::logResponseEnd(req->getTag());
+#ifdef DEBUG
 				AbstractDebugAgent::HttpMessageType msg_type =
 					(&m_rContext.scriptServer() == this)
 					? AbstractDebugAgent::SCRIPT_RESPONSE
 					: AbstractDebugAgent::INCOMING_RESPONSE;
 				DBG_HTTP_MESSAGE_END(msg_type);
-			}
 #endif
+			}
+			
 			if (!body_data.empty() || body_eof) {
 				ScopedSuspender suspender(*this);
-				req->responseHandler().processBodyData(body_data, body_eof, m_downloadProgress);
+				req->responseHandler().processBodyData(body_data, body_eof);
 			}
 		}
 		if (m_pipelineSize == 0 || m_requestQueue.front() != req) {
@@ -706,7 +780,7 @@ Server::processInputData(bool eof)
 				if (errorDescriptor().get()) {
 					answerWithError(errorDescriptor());
 				} else {
-					answerWithError(auto_ptr<ErrorDescriptor>(
+					answerWithError(IntrusivePtr<ErrorDescriptor>(
 						new ErrorDescriptor(
 							ErrorCodes::ERROR_PARSING_RESPONSE,
 							"error parsing response: "+errorMessage(),
@@ -758,13 +832,10 @@ Server::connectionMayBeStored() const
 	if (!m_ptrConnection.get()) {
 		return false;
 	}
-	if (m_ptrConnection->getStatus() != AsyncConnection::ESTABLISHED) {
-		return false;
-	}
 	if (!m_ptrConnection->isExpectedToPersist()) {
 		return false;
 	}
-	if (!m_ptrConnection->getNumResponses() == 0) {
+	if (m_ptrConnection->getNumResponses() == 0) {
 		return false;
 	}
 	if (m_pipelineSize != 0 && m_requestQueue.front()->isConsumingStarted()) {
@@ -827,14 +898,23 @@ Server::restartPendingRequests()
 }
 
 void
-Server::answerWithError(auto_ptr<ErrorDescriptor> edesc)
+Server::answerWithError(IntrusivePtr<ErrorDescriptor> const& edesc)
 {
 	assert(!m_requestQueue.empty());
+	
 	Request& req = *m_requestQueue.front();
+	
+	if (edesc->getErrorCode() == ErrorCodes::REQUEST_CANCEL_TO_SAVE_TRAFFIC) {
+		RequestLog::logResponseEnd(req.getTag());
+	} else {
+		RequestLog::logResponseError(req.getTag());
+	}
+	
 	{
 		ScopedSuspender suspender(*this);
 		req.responseHandler().processError(edesc);
 	}
+	
 	shiftRequestQueue();
 }
 
@@ -862,7 +942,7 @@ void
 Server::onTimeout()
 {
 	URI const& uri = m_requestQueue.front()->getRequestURI();
-	answerWithError(auto_ptr<ErrorDescriptor>(
+	answerWithError(IntrusivePtr<ErrorDescriptor>(
 		new ErrorDescriptor(
 			ErrorCodes::SERVER_TIMEOUT,
 			"server timeout",
@@ -895,7 +975,8 @@ Server::updateState()
 	if (!(m_flags & READING_SUSPENDED)) {
 		if (isArtificialResponseRequired()) {
 			m_delayedActionTimer.ensureActive();
-		} else if (m_pipelineSize > 0 && !m_ptrConnection.get()) {
+		} else if (m_pipelineSize > 0 && !m_ptrConnection.get() &&
+		           !m_connector.isInProgress()) {
 			assert(m_pipelineSize == 1);
 			Request& req = *m_requestQueue.front();
 			if (req.hasData()) {
@@ -904,14 +985,12 @@ Server::updateState()
 		}
 	}
 	
-	if (!isConnectionReady()) {
+	if (!m_ptrConnection.get()) {
 		return; // events and timers are already unregistered / deactivated
 	}
 	
 	if (!(m_flags & WRITE_BUFFER_FULL) && isOutputReady()) {
 		// Necessary on the edge-triggered WFMOReactor.
-		// It must come after requestConnectionFor(),
-		// as it can take a connection from the pool.
 		processOutputData();
 	}
 	
@@ -960,7 +1039,13 @@ Server::processLeadingPhantomResponse()
 	if (!m_phantomRequestQueue.empty()) {
 		if (PhantomRequest* req = m_phantomRequestQueue.front().get()) {
 			auto_ptr<CraftedResponse> response = req->responseGenerator()();
-			response->metadata().setFlag(HttpResponseMetadata::CONTENT_CRAFTED);
+			
+			if (req->isHeadRequest()) {
+				response->metadata().setBodyStatus(HttpMessageMetadata::BODY_FORBIDDEN);
+				response->body().clear();
+			}
+			
+			req->getTag()->flags().set(RequestTag::RESPONSE_CRAFTED);
 			response->metadata().setClientRequestId(req->getClientRequestId());
 #ifdef DEBUG
 			response->metadata().headers().setHeader(
@@ -973,10 +1058,14 @@ Server::processLeadingPhantomResponse()
 				req->responseHandler().processResponseMetadata(
 					response->retrieveMetadata()
 				);
+				
 				req->responseHandler().processBodyData(
-					response->body(), true, DownloadProgress()
+					response->body(), true
 				);
 			}
+			
+			RequestLog::logResponseBegin(req->getTag(), 200);
+			RequestLog::logResponseEnd(req->getTag());
 			m_phantomRequestQueue.pop_front();
 			return true;
 		}
@@ -1080,21 +1169,36 @@ Server::eventGotMetadata(auto_ptr<HttpResponseMetadata> metadata, bool is_persis
 		m_ptrConnection->toBeClosed();
 	}
 	
-	m_downloadProgress.received() = 0;
+	DownloadProgress& dp = req.getTag()->downloadProgress();
+	dp.received() = 0;
 	if (metadata->getBodyStatus() == HttpResponseMetadata::BODY_SIZE_UNKNOWN) {
-		m_downloadProgress.total() = m_downloadProgress.undefined();
+		dp.total() = DownloadProgress::UNDEFINED;
 	} else {
-		m_downloadProgress.total() = metadata->getBodySize();
+		dp.total() = metadata->getBodySize();
 	}
 	
+	RequestLog::logResponseBegin(req.getTag(), metadata->statusLine().getCode());
+	
 	ScopedSuspender suspender(*this);
 	req.responseHandler().processResponseMetadata(metadata);
 }
 
 void
-Server::connectionEstablished()
+Server::connectionEstablished(AutoClosingSAP<ACE_SOCK_Stream>& strm)
 {
 	//DEBUGLOG("Connection established");
+	m_ptrConnection.reset(new ServerConnection(m_connectingTo, strm));
+	m_connectingTo.set(string(), -1);
+	strm.set_handle(ACE_INVALID_HANDLE); // prevent double closing
+	
+	processNewConnection();
+}
+
+void
+Server::processNewConnection()
+{
+	assert(m_ptrConnection.get());
+	
 	m_ptrConnection->setTcpNoDelay();
 	m_ptrConnection->peer().enable(ACE_NONBLOCK);
 	ACE_HANDLE handle = m_ptrConnection->peer().get_handle();
@@ -1112,7 +1216,7 @@ Server::connectionEstablished()
 		DEBUGLOG2(problem);
 		m_ptrConnection.reset(0);
 		Request const& req = *m_requestQueue.front();
-		answerWithError(auto_ptr<ErrorDescriptor>(
+		answerWithError(IntrusivePtr<ErrorDescriptor>(
 			new ErrorDescriptor(
 				ErrorCodes::UNKNOWN_ERROR,
 				problem,
@@ -1140,7 +1244,7 @@ Server::connectionFailed(FailureCode fcode, int errnum)
 	
 	switch (fcode) {
 		case FAIL_DNS: {
-			answerWithError(auto_ptr<ErrorDescriptor>(
+			answerWithError(IntrusivePtr<ErrorDescriptor>(
 				new ErrorDescriptor(
 					ErrorCodes::DNS_ERROR,
 					"could not resolve host name",
@@ -1153,7 +1257,7 @@ Server::connectionFailed(FailureCode fcode, int errnum)
 			break;
 		}
 		case FAIL_TIMEOUT: {
-			answerWithError(auto_ptr<ErrorDescriptor>(
+			answerWithError(IntrusivePtr<ErrorDescriptor>(
 				new ErrorDescriptor(
 					ErrorCodes::CONNECT_TIMEOUT,
 					"connection timeout",
@@ -1172,7 +1276,7 @@ Server::connectionFailed(FailureCode fcode, int errnum)
 			char const* syserr = (fcode == FAIL_REFUSED)
 				? "connection refused"
 				: "destination host unreachable";
-			answerWithError(auto_ptr<ErrorDescriptor>(
+			answerWithError(IntrusivePtr<ErrorDescriptor>(
 				new ErrorDescriptor(
 					ErrorCodes::CONNECT_FAILED,
 					"connection failed",
@@ -1191,7 +1295,7 @@ Server::connectionFailed(FailureCode fcode, int errnum)
 			if (!syserr) {
 				syserr = ACE_OS::strerror(errnum);
 			}
-			answerWithError(auto_ptr<ErrorDescriptor>(
+			answerWithError(IntrusivePtr<ErrorDescriptor>(
 				new ErrorDescriptor(
 					ErrorCodes::CONNECT_FAILED,
 					"connection failed",
@@ -1265,13 +1369,20 @@ Server::onDelayedAction()
 		return;
 	}
 	
+	if (m_pipelineSize > 0 && !m_ptrConnection.get() && !m_connector.isInProgress()) {
+		assert(m_pipelineSize == 1);
+		m_requestQueue.front()->updateRoute();
+		// This may trigger a proxy configuration error, which
+		// will be handled by processArtificialResponses() below.
+	}
+	
 	if (isArtificialResponseRequired()) {
 		processArtificialResponses();
 	}
 	
 	assert(!isArtificialResponseRequired());
 	
-	if (m_pipelineSize > 0 && !m_ptrConnection.get()) {
+	if (m_pipelineSize > 0 && !m_ptrConnection.get() && !m_connector.isInProgress()) {
 		assert(m_pipelineSize == 1);
 		Request& req = *m_requestQueue.front();
 		if (req.hasData()) {
@@ -1286,14 +1397,17 @@ Server::onDelayedAction()
 Server::Request::Request(
 	ServiceContext& context, ServerWeakPtr const& server_ptr,
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& response_handler,
 	auto_ptr<FilterTryList> filter_try_list)
 :	m_rContext(context),
 	m_ptrServer(server_ptr),
 	m_ptrOrigMetadata(request_metadata),
 	m_ptrOutgoingMetadata(new HttpRequestMetadata(*request_metadata)),
+	m_tag(request_tag),
 	m_ptrFilterChain(new ResponseFilterChain(
-		context, m_ptrOrigMetadata, filter_try_list, response_handler, *this
+		context, m_ptrOrigMetadata, request_tag,
+		filter_try_list, response_handler, *this
 	)),
 	m_nextHopAddr("", -1),
 	m_willGoThroughProxy(false),
@@ -1314,7 +1428,46 @@ Server::Request::~Request()
 }
 
 void
-Server::Request::raiseError(auto_ptr<ErrorDescriptor> edesc)
+Server::Request::updateRoute()
+{
+	updateRoute(GlobalState::ReadAccessor()->config());
+}
+
+void
+Server::Request::updateRoute(Config const& conf)
+{
+	URI const& uri = getRequestURI();
+	std::string host = uri.getHost().toStdString();
+	
+	auto_ptr<ProxyDescriptor> proxy;
+	
+	if (!conf.getNextHopProxy().hasProblem()) {
+		proxy = conf.getProxyFor(host);
+	} else {
+		std::string problem = conf.getNextHopProxy().getProblem();
+		setErrorResponse(IntrusivePtr<ErrorDescriptor>(
+			new ErrorDescriptor(
+				ErrorCodes::REMOTE_PROXY_CONFIG_PROBLEM,
+				problem,
+				sigc::bind(
+					sigc::ptr_fun(&ErrorResponse::errRemoteProxyConfig),
+					getRequestURI().toString(), problem
+				)
+			)
+		));
+	}
+	
+	if (proxy.get()) {
+		m_willGoThroughProxy = true;
+		m_nextHopAddr = proxy->getProxyAddr();
+	} else {
+		m_willGoThroughProxy = false;
+		m_nextHopAddr.set(host, uri.guessPort());
+	}
+}
+
+void
+Server::Request::raiseError(IntrusivePtr<ErrorDescriptor> const& edesc)
 {
 	m_ptrErrDesc = edesc;
 }
@@ -1335,7 +1488,7 @@ Server::Request::appendBodyData(SplittableBuffer& data, bool eof)
 	m_streamWriter.appendBodyData(data, eof);
 	if (m_streamWriter.getBufferedBodyDataSize() > MAX_BUFFERED_REQUEST_BODY) {
 		string problem("request body too big");
-		cancelRequest(auto_ptr<ErrorDescriptor>(
+		cancelRequest(IntrusivePtr<ErrorDescriptor>(
 			new ErrorDescriptor(
 				ErrorCodes::REQUEST_BODY_TOO_BIG, problem,
 				sigc::bind(
@@ -1349,7 +1502,7 @@ Server::Request::appendBodyData(SplittableBuffer& data, bool eof)
 }
 
 void
-Server::Request::cancelRequest(auto_ptr<ErrorDescriptor> edesc)
+Server::Request::cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc)
 {
 	setErrorResponse(edesc);
 }
@@ -1382,7 +1535,7 @@ Server::Request::restart()
 	}
 	m_data.clear();
 	m_isConsumingStarted = false;
-	updateRoute(GlobalState::ReadAccessor()->config()); // user settings might have changed
+	updateRoute(); // user settings might have changed
 	m_streamWriter.startRequest(*m_ptrOutgoingMetadata, false, m_willGoThroughProxy, false);
 	SplittableBuffer body_data;
 	m_streamWriter.appendBodyData(body_data, true);
@@ -1411,29 +1564,16 @@ Server::Request::maybeReportClientIP(Config const& conf)
 	}
 }
 
-void
-Server::Request::updateRoute(Config const& conf)
-{
-	URI const& uri = getRequestURI();
-	std::string host = uri.getHost().toStdString();
-	auto_ptr<ProxyDescriptor> proxy = conf.getProxyFor(host);
-	if (proxy.get()) {
-		m_willGoThroughProxy = true;
-		m_nextHopAddr = proxy->getProxyAddr();
-	} else {
-		m_willGoThroughProxy = false;
-		m_nextHopAddr.set(host, uri.guessPort());
-	}
-}
-
 
 /*======================== Server::PhantomRequest ============================*/
 
 Server::PhantomRequest::PhantomRequest(
 	ConstRequestPtr const& request_metadata,
+	RequestTag const& request_tag,
 	IntrusivePtr<AbstractResponseHandler> const& response_handler,
 	CraftedResponseGenerator const& response_generator)
 :	m_ptrOrigMetadata(request_metadata),
+	m_tag(request_tag),
 	m_ptrResponseHandler(response_handler),
 	m_responseGenerator(response_generator)
 {
@@ -1449,6 +1589,6 @@ Server::PhantomRequest::appendBodyData(SplittableBuffer& data, bool eof)
 }
 
 void
-Server::PhantomRequest::cancelRequest(std::auto_ptr<ErrorDescriptor> edesc)
+Server::PhantomRequest::cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc)
 {
 }
