@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,8 +26,9 @@
 
 #include <memory>
 
-class InetAddr;
+class SymbolicInetAddr;
 class ServerConnection;
+class Reactor;
 
 class ServerConnectionPool
 {
@@ -40,10 +41,18 @@ public:
 	
 	virtual void storeConnection(std::auto_ptr<ServerConnection> conn) = 0;
 	
-	virtual std::auto_ptr<ServerConnection> retrieveConnection(InetAddr const& addr) = 0;
+	virtual std::auto_ptr<ServerConnection>
+		retrieveConnection(SymbolicInetAddr const& addr) = 0;
+	
+	// Reactor is used for cleanup purposes. It's optional.
+	// Reactor has to be multi-threaded.
+	virtual void connectToReactor(Reactor& reactor) = 0;
+	
+	virtual void disconnectFromReactor() = 0;
 private:
 	enum { CAPACITY = 60 };
 	enum { TIMEOUT = 12 }; // seconds
+	enum { CLEANUP_TIMEOUT = 5 }; // seconds
 	
 	struct Entry;
 	class Impl;
