@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,12 +28,12 @@
 #include "ConfError.h"
 #include "ConfErrorHandler.h"
 #include "Color.h"
-#include "InetAddr.h"
+#include "SymbolicInetAddr.h"
 #include "ProxyDescriptor.h"
 #include "ArraySize.h"
 #include "StringUtils.h"
 #include "types.h"
-#include <ace/config.h>
+#include <ace/config-lite.h>
 #include <boost/lexical_cast.hpp>
 #include <set>
 #include <string>
@@ -86,7 +86,7 @@ private:
 	
 	bool processListenAddr(
 		char const* begin, char const* end,
-		int line, std::list<InetAddr>& addrs);
+		int line, std::list<SymbolicInetAddr>& addrs);
 	
 	bool handleClientCompression(std::string const& value, int line);
 	
@@ -330,7 +330,7 @@ bool
 ConfigFile::Reader::finalValidation()
 {
 	if (m_rTarget.isNextHopProxyEnabled()) {
-		InetAddr addr = m_rTarget.getNextHopProxy().getProxyAddr();
+		SymbolicInetAddr addr = m_rTarget.getNextHopProxy().getProxyAddr();
 		if (addr.getHost().empty()) {
 			ConfError err(
 				ConfError::T_ERROR,
@@ -420,7 +420,7 @@ ConfigFile::Reader::handleParam(
 bool
 ConfigFile::Reader::handleListenAddress(std::string const& value, int line)
 {
-	list<InetAddr> addrs;
+	list<SymbolicInetAddr> addrs;
 	char const* begin = value.c_str();
 	char const* const end = begin + value.size();
 	char const* rec_end;
@@ -448,7 +448,7 @@ ConfigFile::Reader::handleListenAddress(std::string const& value, int line)
 bool
 ConfigFile::Reader::processListenAddr(
 	char const* begin, char const* end,
-	int line, std::list<InetAddr>& addrs)
+	int line, std::list<SymbolicInetAddr>& addrs)
 {
 	begin = StringUtils::ltrim(begin, end);
 	end = StringUtils::rtrim(begin, end);
@@ -468,7 +468,7 @@ ConfigFile::Reader::processListenAddr(
 		return false;
 	}
 	
-	addrs.push_back(InetAddr(host, port));
+	addrs.push_back(SymbolicInetAddr(host, port));
 	return true;
 }
 
@@ -643,9 +643,9 @@ bool
 ConfigFile::Reader::handleProxyHost(std::string const& value, int line)
 {
 	m_proxyHostLine = line;
-	ProxyDescriptor proxy = m_rTarget.getNextHopProxy();
-	proxy.getProxyAddr().setHost(value);
-	m_rTarget.setNextHopProxy(proxy);
+	SymbolicInetAddr addr(m_rTarget.getNextHopProxy().getProxyAddr());
+	addr.setHost(value);
+	m_rTarget.setNextHopProxy(ProxyDescriptor(addr));
 	return true;
 }
 
@@ -665,9 +665,9 @@ ConfigFile::Reader::handleProxyPort(std::string const& value, int line)
 			return m_rErrorHandler.handleError(err);
 		}
 	}
-	ProxyDescriptor proxy = m_rTarget.getNextHopProxy();
-	proxy.getProxyAddr().setPort(port);
-	m_rTarget.setNextHopProxy(proxy);
+	SymbolicInetAddr addr(m_rTarget.getNextHopProxy().getProxyAddr());
+	addr.setPort(port);
+	m_rTarget.setNextHopProxy(ProxyDescriptor(addr));
 	return true;
 }
 
@@ -900,7 +900,7 @@ ConfigFile::Updater::processParams(
 void
 ConfigFile::Updater::handleListenAddress(Element& el)
 {
-	list<InetAddr> listen_addrs = m_rSource.getListenAddresses();
+	list<SymbolicInetAddr> listen_addrs = m_rSource.getListenAddresses();
 	el.setValue(StringUtils::join(", ", listen_addrs.begin(), listen_addrs.end()));
 }
 
