@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2004  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,13 +28,20 @@
 #include <string>
 
 class ConnAcceptor;
+class ACE_FILE_IO;
 
 class Daemon
 {
 public:
 	bool init(std::string const& confdir);
 	
-	bool run(bool nodaemon);
+	bool run(bool nodaemon, ACE_FILE_IO& pid_file);
+	
+	static int killRunning(ACE_FILE_IO& pid_file);
+	
+	static void initResolver();
+	
+	static std::string normalizePath(std::string const& path);
 private:
 	class ConfigErrorHandler;
 	class ConfigLoader;
