@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,14 +32,16 @@
 #include "BasicConfigDialog.h"
 #include "AdvancedConfigWindow.h"
 #include "FilterConfigWindow.h"
+#include "RequestLogWindow.h"
 #include "Conf.h"
-#include "ConfError.h"
-#include "ConfErrorHandler.h"
+#include "ConfigErrorHandler.h"
 #include "RegexFilterDescriptor.h"
 #include "GlobalState.h"
 #include "UrlPatterns.h"
 #include "FilterGroupTag.h"
 #include "InetAddr.h"
+#include "SymbolicInetAddr.h"
+#include "DnsResolver.h"
 #include "IntrusivePtr.h"
 #include "RefCountableSAP.h"
 #include "AutoClosingSAP.h"
@@ -50,8 +52,8 @@
 #include "DebugWindow.h"
 #endif
 #include "binreloc/prefix.h"
+#include <ace/config-lite.h>
 #include <ace/Dirent.h>
-#include <ace/INET_Addr.h>
 #include <ace/FILE_Addr.h>
 #include <ace/FILE_IO.h>
 #include <ace/FILE_Connector.h>
@@ -69,29 +71,13 @@
 #include <string>
 #include <list>
 #include <deque>
+#include <vector>
 #include <set>
 #include <stddef.h>
 #include <time.h>
 
 using namespace std;
 
-class Application::ConfigErrorHandler : public ConfErrorHandler
-{
-public:
-	ConfigErrorHandler(Glib::ustring const& fname)
-	: m_fileName(fname), m_hasCriticalErrors(false) {}
-	
-	bool hasCriticalErrors() const { return m_hasCriticalErrors; }
-private:
-	virtual bool handleError(ConfError const& err);
-	
-	void writeToLog(ConfError const& err);
-
-	Glib::ustring m_fileName;
-	bool m_hasCriticalErrors;
-};
-
-
 struct Application::ContentFilterFnameComparator
 {
 	bool operator()(
@@ -111,6 +97,7 @@ Application::Application(int argc, char* argv[])
 	m_commandQueue(COMMAND_QUEUE_CAPACITY, *this),
 	m_networkActivityHandler(m_commandQueue),
 	m_filterJsLogHandler(m_commandQueue),
+	m_requestLogHandler(m_commandQueue),
 #ifdef DEBUG
 	m_debugAgent(m_commandQueue),
 #endif
@@ -127,6 +114,7 @@ Application::Application(int argc, char* argv[])
 #ifdef DEBUG
 	m_ptrDebugWindow.reset(new DebugWindow);
 #endif
+	m_ptrAcceptorThread.reset(new AcceptorThread);
 }
 
 Application::~Application()
@@ -140,13 +128,14 @@ Application::run()
 	bool errors = false;
 	errors |= !createMissingUserConfigs();
 	errors |= !loadConfig();
-	errors |= !loadStandardUrlPatterns();
-	errors |= !loadLocalUrlPatterns();
-	errors |= !loadContentFilters();
+	loadStandardUrlPatterns();
+	loadLocalUrlPatterns();
+	loadContentFilters();
 	if (errors) {
 		showLogDialog();
 	}
 	m_gtkMain.run();
+	m_ptrAcceptorThread.reset(0);
 }
 
 void
@@ -197,6 +186,12 @@ Application::showFilterConfigWindow()
 	FilterConfigWindow::showWindow();
 }
 
+void
+Application::showRequestLogWindow()
+{
+	RequestLogWindow::showWindow();
+}
+
 void
 Application::setFilteringEnabled(bool val)
 {
@@ -216,7 +211,7 @@ Application::requestAppExit()
 {
 	// If we just call m_gtkMain.quit(), we may get a deadlock
 	// while waiting for worker threads to finish
-	// (in m_acceptorThread DTOR). This happens if any worker thread
+	// (in AcceptorThread DTOR). This happens if any worker thread
 	// is blocked while sending a command through m_commandQueue.
 	if (m_commandQueue.close() == 0) {
 		// command queue is empty
@@ -414,7 +409,7 @@ Application::processConfig(
 {
 	ConfigErrorHandler eh("config");
 	file_structure.load(text, target, eh);
-	return !eh.hasCriticalErrors();
+	return eh.numErrors() == 0;
 }
 
 bool
@@ -427,18 +422,18 @@ Application::applyConfig(Config const& config)
 
 	GlobalState::WriteAccessor()->config() = config;
 	// We apply the config independently of the result of the
-	// bind operation, to make the GUI will work consistently.
+	// bind operation, to make the Basic Config dialog work as expected.
 	
-	m_acceptorThread.removeAll();
+	m_ptrAcceptorThread->removeAll();
 	setAcceptingConnections(false);
 	
 	bool ok = true;
 	deque<AcceptorPtr> acceptors;
-	list<InetAddr> addrs = config.getListenAddresses();
+	list<SymbolicInetAddr> addrs = config.getListenAddresses();
 	for (; !addrs.empty(); addrs.pop_front()) {
-		InetAddr const& addr = addrs.front();
-		ACE_INET_Addr resolved_addr;
-		if (!addr.resolve(resolved_addr)) {
+		SymbolicInetAddr const& addr = addrs.front();
+		vector<InetAddr> resolved_addrs = DnsResolver::resolve(addr);
+		if (resolved_addrs.empty()) {
 			ostringstream err;
 			err << "Could not resolve listen address \"" << addr << '"';
 			log->appendRecord(err.str().c_str(), log->getErrorStyle());
@@ -446,7 +441,7 @@ Application::applyConfig(Config const& config)
 			continue;
 		}
 		AcceptorPtr acceptor(new Acceptor);
-		if (acceptor->open(resolved_addr, true) == -1) {
+		if (acceptor->open(resolved_addrs[0], true) == -1) {
 			ostringstream err;
 			err << "Could not bind to \"" << addr << '"';
 			log->appendRecord(err.str().c_str(), log->getErrorStyle());
@@ -467,29 +462,41 @@ Application::applyConfig(Config const& config)
 	}
 	
 	for (; !acceptors.empty(); acceptors.pop_front()) {
-		m_acceptorThread.add(acceptors.front());
+		m_ptrAcceptorThread->add(acceptors.front());
 	}
 	setAcceptingConnections(true);
 	
 	return true;
 }
 
+void
+Application::applyBrokenConfig(Config const& broken_config)
+{
+	GlobalState::WriteAccessor()->config() = broken_config;
+	// We apply it anyway, to make the Basic Config dialog work as expected.
+	
+	m_ptrAcceptorThread->removeAll();
+	setAcceptingConnections(false);
+}
+
 bool
 Application::loadConfig()
 {
 	Log* log = OperationLog::instance();
 	log->appendRecord("Loading config ... ");
 	size_t num_records = log->getNumRecords();
-
+	
 	string const abs_fname = Glib::build_filename(m_userConfDir, "config");
 	string text;
 	time_t mtime = 0;
 	if (!readFile(abs_fname, text, &mtime)) {
 		return false;
 	}
+	EffectiveFileTimestamps::config = mtime;
 	
 	Config config;
 	if (!processConfig(text, config, m_configFile)) {
+		applyBrokenConfig(config);
 		return false;
 	}
 	
@@ -497,8 +504,6 @@ Application::loadConfig()
 		return false;
 	}
 	
-	EffectiveFileTimestamps::config = mtime;
-	
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord("done", log->getSuccessStyle());
 	}
@@ -513,7 +518,7 @@ Application::processUrlPatterns(
 {
 	ConfigErrorHandler eh(filenameToUtf8(fname));
 	file_structure.load(text, target, eh);
-	return !eh.hasCriticalErrors();
+	return eh.numErrors() == 0;
 }
 
 void
@@ -528,7 +533,7 @@ Application::applyLocalUrlPatterns(UrlPatterns const& patterns)
 	GlobalState::WriteAccessor()->urlPatterns().localPatterns() = patterns;
 }
 
-bool
+void
 Application::loadStandardUrlPatterns()
 {
 	Log* log = OperationLog::instance();
@@ -538,25 +543,18 @@ Application::loadStandardUrlPatterns()
 	string const fname("urls");
 	string const abs_fname = Glib::build_filename(m_globalConfDir, fname);
 	string text;
-	if (!readFile(abs_fname, text)) {
-		return false;
-	}
+	readFile(abs_fname, text);
 	
 	UrlPatterns patterns;
-	if (!processUrlPatterns(text, fname,
-	                        patterns, m_standardUrlPatternsFile)) {
-		return false;
-	}
-	
+	processUrlPatterns(text, fname, patterns, m_standardUrlPatternsFile);
 	applyStandardUrlPatterns(patterns);
 	
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord("done", log->getSuccessStyle());
 	}
-	return true;
 }
 
-bool
+void
 Application::loadLocalUrlPatterns()
 {
 	Log* log = OperationLog::instance();
@@ -567,16 +565,10 @@ Application::loadLocalUrlPatterns()
 	string const abs_fname = Glib::build_filename(m_userConfDir, fname);
 	string text;
 	time_t mtime = 0;
-	if (!readFile(abs_fname, text, &mtime)) {
-		return false;
-	}
+	readFile(abs_fname, text, &mtime);
 	
 	UrlPatterns patterns;
-	if (!processUrlPatterns(text, fname,
-	                        patterns, m_localUrlPatternsFile)) {
-		return false;
-	}
-	
+	processUrlPatterns(text, fname, patterns, m_localUrlPatternsFile);
 	applyLocalUrlPatterns(patterns);
 	
 	EffectiveFileTimestamps::urls_local = mtime;
@@ -584,10 +576,9 @@ Application::loadLocalUrlPatterns()
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord("done", log->getSuccessStyle());
 	}
-	return true;
 }
 
-bool
+void
 Application::loadContentFilters()
 {
 	Log* log = OperationLog::instance();
@@ -599,18 +590,13 @@ Application::loadContentFilters()
 	deque<string> global_fnames;
 	deque<string> user_fnames;
 	
-	if (!loadFilterFnames(global_dir, global_fnames)) {
-		return false;
-	}
-	if (!loadFilterFnames(user_dir, user_fnames)) {
-		return false;
-	}
+	loadFilterFnames(global_dir, global_fnames);
+	loadFilterFnames(user_dir, user_fnames);
 	
 	std::sort(global_fnames.begin(), global_fnames.end());
 	std::sort(user_fnames.begin(), user_fnames.end());
 	
 	m_contentFilters.clear();
-	bool critical_errors = false;	
 	
 	while (true) {
 		ContentFilterGroup::FileLocation location;
@@ -638,13 +624,12 @@ Application::loadContentFilters()
 		string text;
 		string abs_fname = Glib::build_filename(dirname, fname);
 		if (!readFile(abs_fname, text)) {
-			critical_errors = true;
 			continue;
 		}
 		
 		IntrusivePtr<ContentFilterGroup> group(
 			new ContentFilterGroup(
-				FilterGroupTag::create(), fname, location
+				FilterGroupTag(), fname, location
 			)
 		);
 		
@@ -652,9 +637,7 @@ Application::loadContentFilters()
 		group->fileStructure().load(
 			text, group->filters(), eh, group->getTag()
 		);
-		if (eh.hasCriticalErrors()) {
-			critical_errors = true;
-		} else {
+		if (eh.numErrors() == 0) {
 			loadEnabledFilterList(*group, user_dir);
 		}
 		
@@ -666,7 +649,6 @@ Application::loadContentFilters()
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord("done", log->getSuccessStyle());
 	}
-	return !critical_errors;
 }
 
 bool
@@ -801,6 +783,7 @@ Application::notifyCommandsPending()
 	m_dispatcher.emit();
 }
 
+// Note: this function is called from the main thread only.
 void
 Application::notifyQueueEmpty()
 {
@@ -974,37 +957,3 @@ Application::setAcceptingConnections(bool val)
 		m_acceptingConnectionsSignal.emit(val);
 	}
 }
-
-
-/*=================== Application::ConfigErrorHandler ====================*/
-
-bool
-Application::ConfigErrorHandler::handleError(ConfError const& err)
-{
-	if (err.getType() == err.T_ERROR) {
-		m_hasCriticalErrors = true;
-	}
-	writeToLog(err);
-	return true;
-}
-
-void
-Application::ConfigErrorHandler::writeToLog(ConfError const& err)
-{
-	ostringstream text;
-	text << '[' << m_fileName.raw();
-	if (err.getLine()) {
-		text << ':' << err.getLine();
-		if (err.getCol()) {
-			text << ':' << err.getCol();
-		}
-	}
-	text << "] " << localeToUtf8(err.getMessage()).raw();
-	Log* log = OperationLog::instance();
-	log->appendRecord(
-		text.str(),
-		err.getType() == err.T_WARNING
-		? log->getWarningStyle()
-		: log->getErrorStyle()
-	);
-}
