@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -41,6 +41,7 @@
 #include "ScopedIncDec.h"
 #include "StringUtils.h"
 #include "CompiledImages.h"
+#include <ace/config-lite.h>
 #include <ace/OS_NS_unistd.h>
 #include <boost/lexical_cast.hpp>
 #include <boost/regex.hpp>
@@ -58,7 +59,9 @@
 #include <gtkmm/alignment.h>
 #include <gtkmm/frame.h>
 #include <gtkmm/scrolledwindow.h>
+#include <gtkmm/accelkey.h>
 #include <gtkmm/menu.h>
+#include <gtkmm/menubar.h>
 #include <gtkmm/image.h>
 #include <gtkmm/stock.h>
 #include <gtkmm/messagedialog.h>
@@ -205,6 +208,8 @@ public:
 	
 	static Glib::RefPtr<FilterTreeModel> create();
 	
+	void reload();
+	
 	FilterTreeColumns const& getCols() const { return m_columns; }
 	
 	sigc::slot<void, Glib::ustring const&> getCheckBoxHandler();
@@ -213,11 +218,11 @@ public:
 	
 	bool isFilterItem(Row const& row) const;
 	
+	bool isDeleting() const { return m_isDeleting != 0; }
+	
 	iterator erase(iterator const& iter);
 	
-	sigc::signal<void, Gtk::TreeIter const&>& signalItemBeingDeleted() {
-		return m_itemBeingDeletedSignal;
-	}
+	void clear();
 	
 	void updateFileCheckState(Row const& file_row);
 	
@@ -254,7 +259,7 @@ private:
 		Row const& row, CheckState old_state, CheckState new_state);
 	
 	FilterTreeColumns m_columns;
-	sigc::signal<void, Gtk::TreeIter const&> m_itemBeingDeletedSignal;
+	int m_isDeleting;
 };
 
 
@@ -304,8 +309,6 @@ private:
 	void onEditingFinished(
 		Glib::ustring const& old_text, Glib::ustring const& new_text);
 	
-	void onItemBeingDeleted(Gtk::TreeIter const& item);
-	
 	void onSelChanged();
 	
 	FilterConfigWindow& m_rOwner;
@@ -388,6 +391,8 @@ FilterConfigWindow::FilterConfigWindow()
 	Gtk::VBox* top_vbox = manage(new Gtk::VBox);
 	add(*top_vbox);
 	
+	top_vbox->pack_start(*manage(createMenuBar().release()), Gtk::PACK_SHRINK);
+	
 	Gtk::HBox* lr_box = manage(new Gtk::HBox);
 	top_vbox->pack_start(*lr_box);
 	
@@ -482,6 +487,39 @@ FilterConfigWindow::on_delete_event(GdkEventAny*)
 	return true; // stop event propagation
 }
 
+std::auto_ptr<Gtk::MenuBar>
+FilterConfigWindow::createMenuBar()
+{
+	std::auto_ptr<Gtk::Menu> file(new Gtk::Menu());
+	file->items().push_back(
+		Gtk::Menu_Helpers::ImageMenuElem(
+			"New",
+			Gtk::AccelKey("<control>n"),
+			*manage(new Gtk::Image(getFileIcon())),
+			sigc::mem_fun(*this, &FilterConfigWindow::onNewFile)
+		)
+	);
+	
+	file->items().push_back(
+		Gtk::Menu_Helpers::ImageMenuElem(
+			"Reload",
+			Gtk::AccelKey("<control>r"),
+			*manage(new Gtk::Image(Gtk::Stock::REFRESH, Gtk::ICON_SIZE_MENU)),
+			sigc::mem_fun(*this, &FilterConfigWindow::onReload)
+		)
+	);
+	
+	std::auto_ptr<Gtk::MenuBar> mbar(new Gtk::MenuBar());
+	mbar->items().push_back(
+		Gtk::Menu_Helpers::MenuElem(
+			"_File",
+			*manage(file.release())
+		)
+	);
+	
+	return mbar;
+}
+
 bool
 FilterConfigWindow::prepareForWindowDestruction()
 {
@@ -563,7 +601,7 @@ FilterConfigWindow::onSave()
 		new_filter.reset(new RegexFilterDescriptor(*node->filter));
 	} else {
 		new_filter.reset(new RegexFilterDescriptor(
-			FilterTag::create(),
+			FilterTag(),
 			file_node->filterGroup->getTag(),
 			Application::localeFromUtf8(row[getCols().NAME])
 		));
@@ -674,7 +712,7 @@ FilterConfigWindow::createFile(
 	
 	IntrusivePtr<ContentFilterGroup> group(
 		new ContentFilterGroup(
-			FilterGroupTag::create(), fname,
+			FilterGroupTag(), fname,
 			ContentFilterGroup::USER_DIR
 		)
 	);
@@ -936,6 +974,34 @@ FilterConfigWindow::onItemSelected(Gtk::TreeIter const& item)
 	}
 }
 
+void
+FilterConfigWindow::onNewFile()
+{
+	m_pFilterTree->onNewFile();
+}
+
+void
+FilterConfigWindow::onReload()
+{
+	if (m_ptrFilterTreeModel->haveModifiedItems()) {
+		Gtk::MessageDialog dialog(
+			*this,
+			"Reload filters and lose <b>unsaved changes</b>?",
+			true, /* use_markup */
+			Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO,
+			true /* modal */
+		);
+		int const response = dialog.run();
+		if (response != Gtk::RESPONSE_YES) {
+			return;
+		}
+	}
+	
+	Application::instance()->loadContentFilters();
+	m_ptrFilterTreeModel->reload();
+	// FilterPanel will update itself
+}
+
 void
 FilterConfigWindow::showError(Glib::ustring const& msg, bool markup)
 {
@@ -1096,6 +1162,7 @@ FilterConfigWindow::MyCellRendererToggle::activate_vfunc(
 /*================= FilterConfigWindow::FilterTreeModel ==================*/
 
 FilterConfigWindow::FilterTreeModel::FilterTreeModel()
+:	m_isDeleting(0)
 {
 	set_column_types(m_columns);
 	loadFilterGroups();
@@ -1111,6 +1178,13 @@ FilterConfigWindow::FilterTreeModel::create()
 	return Glib::RefPtr<FilterTreeModel>(new FilterTreeModel);
 }
 
+void
+FilterConfigWindow::FilterTreeModel::reload()
+{
+	clear();
+	loadFilterGroups();
+}
+
 sigc::slot<void, Glib::ustring const&>
 FilterConfigWindow::FilterTreeModel::getCheckBoxHandler()
 {
@@ -1182,13 +1256,20 @@ FilterConfigWindow::FilterTreeModel::isFilterItem(Row const& row) const
 	return row.parent();
 }
 
-Gtk::TreeStore::iterator
+Gtk::TreeIter
 FilterConfigWindow::FilterTreeModel::erase(iterator const& iter)
 {
-	m_itemBeingDeletedSignal.emit(iter);
+	ScopedIncDec<int> incrementer(m_isDeleting);
 	return Gtk::TreeStore::erase(iter);
 }
 
+void
+FilterConfigWindow::FilterTreeModel::clear()
+{
+	ScopedIncDec<int> incrementer(m_isDeleting);
+	Gtk::TreeStore::clear();
+}
+
 void
 FilterConfigWindow::FilterTreeModel::updateFileCheckState(Row const& file_row)
 {
@@ -1487,9 +1568,6 @@ FilterConfigWindow::FilterTree::FilterTree(
 	
 	append_column(m_column);
 	
-	m_ptrModel->signalItemBeingDeleted().connect(
-		sigc::mem_fun(*this, &FilterTree::onItemBeingDeleted)
-	);
 	get_selection()->signal_changed().connect(
 		sigc::mem_fun(*this, &FilterTree::onSelChanged)
 	);
@@ -1852,19 +1930,21 @@ FilterConfigWindow::FilterTree::onEditingFinished(
 	m_rOwner.handleItemRename(sel, new_text);
 }
 
-void
-FilterConfigWindow::FilterTree::onItemBeingDeleted(Gtk::TreeIter const& item)
-{
-	if (item == m_selectedItem) {
-		//m_rOwner.onItemUnselected(item); // not really necessary
-		m_selectedItem = m_ptrModel->children().end();
-	}
-}
-
 void
 FilterConfigWindow::FilterTree::onSelChanged()
 {
-	m_rOwner.onItemUnselected(m_selectedItem);
+	if (m_ptrModel->isDeleting()) {
+		/*
+		When a selected row is being deleted, the selection_changed
+		signal is emitted after the node has been removed from the
+		model, but before row_deleted signal is emitted.
+		If we get here, m_selectedItem is invalid. Fortunately, we
+		don't really need to call onItemUnselected() in this case.
+		*/
+	} else {
+		m_rOwner.onItemUnselected(m_selectedItem);
+	}
+	
 	m_selectedItem = get_selection()->get_selected();
 	m_rOwner.onItemSelected(m_selectedItem);
 }
