@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,10 +29,9 @@
 
 class InterthreadCommandQueue;
 
-class NetworkActivityHandler :
-	public AbstractNetworkActivityHandler,
-	private NonCopyable
+class NetworkActivityHandler : public AbstractNetworkActivityHandler
 {
+	DECLARE_NON_COPYABLE(NetworkActivityHandler)
 public:
 	NetworkActivityHandler(InterthreadCommandQueue& command_queue);
 	
