@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,8 +29,9 @@
 #include <gtkmm/window.h>
 #include <memory>
 
-class StatusDocklet : public sigc::trackable, private NonCopyable
+class StatusDocklet : public sigc::trackable
 {
+	DECLARE_NON_COPYABLE(StatusDocklet)
 public:
 	StatusDocklet();
 	
