@@ -39,7 +39,9 @@ protected:
 	
 	void onFilterConfig();
 	
-	void onLog();
+	void onOperationLog();
+	
+	void onRequestLog();
 #ifdef DEBUG
 	void onDebug();
 #endif
