@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,25 +28,28 @@
 #include "TrayIcon.h"
 #include "TrayMenu.h"
 #include "LogDialog.h"
+#include "RequestLogWindow.h"
 #include "BasicConfigDialog.h"
 #include "AdvancedConfigWindow.h"
 #include "FilterConfigWindow.h"
 #include "Conf.h"
-#include "ConfError.h"
-#include "ConfErrorHandler.h"
+#include "ConfigErrorHandler.h"
 #include "RegexFilterDescriptor.h"
 #include "FilterGroupTag.h"
 #include "GlobalState.h"
 #include "UrlPatterns.h"
 #include "InetAddr.h"
+#include "SymbolicInetAddr.h"
+#include "DnsResolver.h"
 #include "IntrusivePtr.h"
 #include "RefCountableSAP.h"
 #include "AutoClosingSAP.h"
 #include "AcceptorThread.h"
 #include "ScopedIncDec.h"
 #include "EffectiveFileTimestamps.h"
+#include "AtomicOps.h"
+#include <ace/config-lite.h>
 #include <ace/Dirent.h>
-#include <ace/INET_Addr.h>
 #include <ace/FILE_Addr.h>
 #include <ace/FILE_IO.h>
 #include <ace/FILE_Connector.h>
@@ -66,29 +69,13 @@
 #include <string>
 #include <list>
 #include <deque>
+#include <vector>
 #include <set>
 #include <algorithm>
 #include <stddef.h>
 
 using namespace std;
 
-class Application::ConfigErrorHandler : public ConfErrorHandler
-{
-public:
-	ConfigErrorHandler(wxString const& fname)
-	: m_fileName(fname), m_hasCriticalErrors(false) {}
-	
-	bool hasCriticalErrors() const { return m_hasCriticalErrors; }
-private:
-	virtual bool handleError(ConfError const& err);
-	
-	void writeToLog(ConfError const& err);
-
-	wxString m_fileName;
-	bool m_hasCriticalErrors;
-};
-
-
 struct Application::ContentFilterFnameComparator
 {
 	bool operator()(
@@ -101,22 +88,19 @@ struct Application::ContentFilterFnameComparator
 
 IMPLEMENT_APP_NO_MAIN(Application)
 
-BEGIN_DECLARE_EVENT_TYPES()
-	DECLARE_LOCAL_EVENT_TYPE(myEVT_COMMANDS_PENDING, 0)
-END_DECLARE_EVENT_TYPES()
-DEFINE_EVENT_TYPE(myEVT_COMMANDS_PENDING)
-
 BEGIN_EVENT_TABLE(Application, wxApp)
-	EVT_COMMAND(wxID_ANY, myEVT_COMMANDS_PENDING, onCommandsPending)
+	EVT_IDLE(onIdle)
 END_EVENT_TABLE()
 
 
 Application::Application()
 :	m_isExiting(false),
 	m_commandNestLevel(0),
+	m_havePendingCommands(0),
 	m_commandQueue(COMMAND_QUEUE_CAPACITY, *this),
 	m_networkActivityHandler(m_commandQueue),
 	m_filterJsLogHandler(m_commandQueue),
+	m_requestLogHandler(m_commandQueue),
 	m_isAcceptingConnections(false)
 {
 }
@@ -136,12 +120,13 @@ Application::OnInit()
 	m_confDir.AppendDir(_T("conf"));
 	m_ptrTrayIcon.reset(new TrayIcon);
 	m_ptrTrayMenu.reset(new TrayMenu);
+	m_ptrAcceptorThread.reset(new AcceptorThread);
 	
 	bool errors = false;
 	errors |= !loadConfig();
-	errors |= !loadStandardUrlPatterns();
-	errors |= !loadLocalUrlPatterns();
-	errors |= !loadContentFilters();
+	loadStandardUrlPatterns();
+	loadLocalUrlPatterns();
+	loadContentFilters();
 	if (errors) {
 		showLogDialog();
 	}
@@ -151,6 +136,9 @@ Application::OnInit()
 int
 Application::OnExit()
 {
+	// This will also finish the worker threads.
+	m_ptrAcceptorThread.reset(0);
+	
 	/*
 	Initialyzing these things in the constructor
 	or deleting them in the destructor makes WxWidgets crash.
@@ -173,6 +161,12 @@ Application::showLogDialog()
 	LogDialog::show();
 }
 
+void
+Application::showRequestLogWindow()
+{
+	RequestLogWindow::show();
+}
+
 void
 Application::showBasicConfigDialog()
 {
@@ -202,7 +196,7 @@ Application::requestAppExit()
 {
 	// If we just call ExitMainLoop(), we may get a deadlock
 	// while waiting for worker threads to finish
-	// (in m_acceptorThread DTOR). This happens if any worker thread
+	// (in AcceptorThread DTOR). This happens if any worker thread
 	// is blocked while sending a command through m_commandQueue.
 	if (m_commandQueue.close() == 0) {
 		// command queue is empty
@@ -355,7 +349,7 @@ Application::processConfig(
 {
 	ConfigErrorHandler eh(_T("config"));
 	file_structure.load(text, target, eh);
-	return !eh.hasCriticalErrors();
+	return eh.numErrors() == 0;
 }
 
 bool
@@ -368,18 +362,18 @@ Application::applyConfig(Config const& config)
 
 	GlobalState::WriteAccessor()->config() = config;
 	// We apply the config independently of the result of the
-	// bind operation, to make the GUI will work consistently.
+	// bind operation, to make the Basic Config dialog work as expected.
 	
-	m_acceptorThread.removeAll();
+	m_ptrAcceptorThread->removeAll();
 	setAcceptingConnections(false);
 	
 	bool ok = true;
 	deque<AcceptorPtr> acceptors;
-	list<InetAddr> addrs = config.getListenAddresses();
+	list<SymbolicInetAddr> addrs = config.getListenAddresses();
 	for (; !addrs.empty(); addrs.pop_front()) {
-		InetAddr const& addr = addrs.front();
-		ACE_INET_Addr resolved_addr;
-		if (!addr.resolve(resolved_addr)) {
+		SymbolicInetAddr const& addr = addrs.front();
+		vector<InetAddr> resolved_addrs = DnsResolver::resolve(addr);
+		if (resolved_addrs.empty()) {
 			ostringstream err;
 			err << "Could not resolve listen address \"" << addr << '"';
 			log->appendRecord(err.str().c_str(), log->getErrorStyle());
@@ -387,7 +381,7 @@ Application::applyConfig(Config const& config)
 			continue;
 		}
 		AcceptorPtr acceptor(new Acceptor);
-		if (acceptor->open(resolved_addr, true) == -1) {
+		if (acceptor->open(resolved_addrs[0], true) == -1) {
 			ostringstream err;
 			err << "Could not bind to \"" << addr << '"';
 			log->appendRecord(err.str().c_str(), log->getErrorStyle());
@@ -408,13 +402,23 @@ Application::applyConfig(Config const& config)
 	}
 	
 	for (; !acceptors.empty(); acceptors.pop_front()) {
-		m_acceptorThread.add(acceptors.front());
+		m_ptrAcceptorThread->add(acceptors.front());
 	}
 	setAcceptingConnections(true);
 	
 	return true;
 }
 
+void
+Application::applyBrokenConfig(Config const& broken_config)
+{
+	GlobalState::WriteAccessor()->config() = broken_config;
+	// We apply it anyway, to make the Basic Config dialog work as expected.
+	
+	m_ptrAcceptorThread->removeAll();
+	setAcceptingConnections(false);
+}
+
 bool
 Application::loadConfig()
 {
@@ -429,9 +433,11 @@ Application::loadConfig()
 	if (!readFile(fname, text, &mtime)) {
 		return false;
 	}
+	EffectiveFileTimestamps::config = mtime;
 	
 	Config config;
 	if (!processConfig(text, config, m_configFile)) {
+		applyBrokenConfig(config);
 		return false;
 	}
 	
@@ -439,8 +445,6 @@ Application::loadConfig()
 		return false;
 	}
 	
-	EffectiveFileTimestamps::config = mtime;
-	
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
 	}
@@ -455,7 +459,7 @@ Application::processUrlPatterns(
 {
 	ConfigErrorHandler eh(fname);
 	file_structure.load(text, target, eh);
-	return !eh.hasCriticalErrors();
+	return eh.numErrors() == 0;
 }
 
 void
@@ -470,7 +474,7 @@ Application::applyLocalUrlPatterns(UrlPatterns const& patterns)
 	GlobalState::WriteAccessor()->urlPatterns().localPatterns() = patterns;
 }
 
-bool
+void
 Application::loadStandardUrlPatterns()
 {
 	Log* log = OperationLog::instance();
@@ -480,25 +484,21 @@ Application::loadStandardUrlPatterns()
 	wxFileName fname(m_confDir);
 	fname.SetName(_T("urls"));
 	string text;
-	if (!readFile(fname, text)) {
-		return false;
-	}
+	readFile(fname, text);
 	
 	UrlPatterns patterns;
-	if (!processUrlPatterns(text, fname.GetFullName(),
-	                        patterns, m_standardUrlPatternsFile)) {
-		return false;
-	}
-	
+	processUrlPatterns(
+		text, fname.GetFullName(),
+		patterns, m_standardUrlPatternsFile
+	);
 	applyStandardUrlPatterns(patterns);
 	
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
 	}
-	return true;
 }
 
-bool
+void
 Application::loadLocalUrlPatterns()
 {
 	Log* log = OperationLog::instance();
@@ -509,16 +509,13 @@ Application::loadLocalUrlPatterns()
 	fname.SetName(_T("urls.local"));
 	string text;
 	time_t mtime = 0;
-	if (!readFile(fname, text, &mtime)) {
-		return false;
-	}
+	readFile(fname, text, &mtime);
 	
 	UrlPatterns patterns;
-	if (!processUrlPatterns(text, fname.GetFullName(),
-	                        patterns, m_localUrlPatternsFile)) {
-		return false;
-	}
-	
+	processUrlPatterns(
+		text, fname.GetFullName(),
+		patterns, m_localUrlPatternsFile
+	);
 	applyLocalUrlPatterns(patterns);
 	
 	EffectiveFileTimestamps::urls_local = mtime;
@@ -526,10 +523,9 @@ Application::loadLocalUrlPatterns()
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
 	}
-	return true;
 }
 
-bool
+void
 Application::loadContentFilters()
 {
 	Log* log = OperationLog::instance();
@@ -537,33 +533,12 @@ Application::loadContentFilters()
 	size_t num_records = log->getNumRecords();
 	
 	wxFileName dirname = getFiltersDir();
-	wxString const dirname_str = dirname.GetFullPath();
-	
-	ACE_Dirent dir;
-	if (dir.open(dirname_str.c_str()) == -1) {
-		log->appendRecord(
-			_T("Could not open directory ") + dirname_str,
-			log->getErrorStyle()
-		);
-		return false;
-	}
-	
 	deque<wxString> fnames;
-	
-	for (dirent* ent; (ent = dir.read()); ) {
-		wxString const fname(ent->d_name);
-		if (fname.Find(_T('.')) != -1) {
-			// skip files with extensions
-			continue;
-		}
-		fnames.push_back(fname);
-	}
+	loadFilterFnames(dirname, fnames);
 	
 	std::sort(fnames.begin(), fnames.end());
 	m_contentFilters.clear();
 	
-	bool critical_errors = false;	
-	
 	for (; !fnames.empty(); fnames.pop_front()) {
 		wxString fname(fnames.front());
 		wxFileName abs_fname(dirname);
@@ -571,21 +546,18 @@ Application::loadContentFilters()
 		
 		string text;
 		if (!readFile(abs_fname, text)) {
-			critical_errors = true;
 			continue;
 		}
 		
 		IntrusivePtr<ContentFilterGroup> group(
-			new ContentFilterGroup(FilterGroupTag::create(), fname)
+			new ContentFilterGroup(FilterGroupTag(), fname)
 		);
 		
 		ConfigErrorHandler eh(fname);
 		group->fileStructure().load(
 			text, group->filters(), eh, group->getTag()
 		);
-		if (eh.hasCriticalErrors()) {
-			critical_errors = true;
-		} else {
+		if (eh.numErrors() == 0) {
 			loadEnabledFilterList(*group, dirname);
 		}
 		
@@ -597,7 +569,34 @@ Application::loadContentFilters()
 	if (num_records == log->getNumRecords()) {
 		log->appendToLastRecord(_T("done"), log->getSuccessStyle());
 	}
-	return !critical_errors;
+}
+
+bool
+Application::loadFilterFnames(
+	wxFileName const& dirname, std::deque<wxString>& fnames)
+{
+	Log* log = OperationLog::instance();
+	wxString dirname_str = dirname.GetFullPath();
+	
+	ACE_Dirent dir;
+	if (dir.open(dirname_str.c_str()) == -1) {
+		log->appendRecord(
+			_T("Could not open directory ") + dirname_str,
+			log->getErrorStyle()
+		);
+		return false;
+	}
+	
+	for (dirent* ent; (ent = dir.read()); ) {
+		wxString const fname(ent->d_name);
+		if (fname.Find(_T('.')) != -1) {
+			// skip files with extensions
+			continue;
+		}
+		fnames.push_back(fname);
+	}
+	
+	return true;
 }
 
 void
@@ -743,9 +742,11 @@ Application::handleNetworkActivity()
 void
 Application::notifyCommandsPending()
 {
-	AddPendingEvent(wxCommandEvent(myEVT_COMMANDS_PENDING));
+	AtomicOps::set(&m_havePendingCommands, 1);
+	wxWakeUpIdle();
 }
 
+// Note: this function is called from the main thread only.
 void
 Application::notifyQueueEmpty()
 {
@@ -755,7 +756,7 @@ Application::notifyQueueEmpty()
 }
 
 void
-Application::onCommandsPending(wxCommandEvent& evt)
+Application::onIdle(wxIdleEvent& evt)
 {
 	processCommands();
 }
@@ -769,6 +770,10 @@ Application::processCommands()
 		return;
 	}
 	
+	if (AtomicOps::set(&m_havePendingCommands, 0) == 0) {
+		return;
+	}
+	
 	ScopedIncDec<int> nest_level_manager(m_commandNestLevel);
 	
 	unsigned max_commands = m_commandQueue.getCapacity();
@@ -794,36 +799,3 @@ Application::setAcceptingConnections(bool val)
 		m_acceptingConnectionsSignal.emit(val);
 	}
 }
-
-/*=================== Application::ConfigErrorHandler ====================*/
-
-bool
-Application::ConfigErrorHandler::handleError(ConfError const& err)
-{
-	if (err.getType() == err.T_ERROR) {
-		m_hasCriticalErrors = true;
-	}
-	writeToLog(err);
-	return true;	
-}
-
-void
-Application::ConfigErrorHandler::writeToLog(ConfError const& err)
-{
-	wxString text;
-	text << _T("[") << m_fileName;
-	if (err.getLine()) {
-		text << _T(":") << err.getLine();
-		if (err.getCol()) {
-			text << _T(":") << err.getCol();
-		}
-	}
-	text << _T("] ") << err.getMessage().c_str();
-	Log* log = OperationLog::instance();
-	log->appendRecord(
-		text,
-		err.getType() == err.T_WARNING
-		? log->getWarningStyle()
-		: log->getErrorStyle()
-	);
-}
