@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,14 +25,15 @@
 #endif
 
 #include "AbstractLogView.h"
-#include <wx/dialog.h>
+#include <ace/config-lite.h>
+#include <wx/frame.h>
 
 class LogWidget;
 class wxActivateEvent;
 class wxCloseEvent;
 class wxCommandEvent;
 
-class LogDialog : public wxDialog, private AbstractLogView
+class LogDialog : public wxFrame, private AbstractLogView
 {
 public:
 	LogDialog();
@@ -45,6 +46,8 @@ private:
 	
 	void onWindowClose(wxCloseEvent& evt);
 	
+	void onClose(wxCommandEvent& evt);
+
 	void onClearLog(wxCommandEvent& evt);
 	
 	static LogDialog* m_spInstance;
