@@ -1,6 +1,6 @@
 /*
     BFilter - a smart ad-filtering web proxy
-    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>
+    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
 #define TEXTSTYLE_H_
 
 #include "AtomicCounter.h"
-#include <ace/config.h>
+#include <ace/config-lite.h>
 #include <wx/textctrl.h>
 
 class TextStyle : public wxTextAttr
